#include <string>
#include <set>
#include <map>
#include <fstream>
#include "ConfigParser/ConfigParser.h"
#include "Descriptor/ShapeDescriptors.h"
#include "Utils/Utils.h"
using namespace std;
static char common_setting_filename[] = "configs/config.cfg";
static char specific_setting_filename[] = "configs/gen_shape_descriptors.cfg";
static int
ParseArgs(int argc, char **argv)
{
    // Parse arguments
    argc--; argv++;
    while (argc > 0) {
        if ((*argv)[0] == '-') {
			if (!strcmp(*argv, "-config"))
			{
				argv++; argc--;
				MyConfigParser.Import(*argv);
			}
            else if (!strcmp(*argv, "-Help")) {
                MyConfigParser.PrintAllSettings();
                exit(-1);
            }
			else {
				if (MyConfigParser.HasSection((*argv)+1)) {
					char* sec = (*argv) + 1;
					argv++; argc--;
					char* opt = *argv;
					argv++; argc--;
					char* val = *argv;
					MyConfigParser.Set(sec, opt, val);
				}
				else {
					fprintf(stderr, "Invalid program argument: %s", (*argv)+1);
					exit(1);
				}
			}
            argv++; argc--;
        }
        else {
            fprintf(stderr, "Invalid program argument: %s", *argv); exit(1);
            argv++; argc--;
        }
    }
    
    // Return OK status
    return 1;
}

static int
GenerateOneArff(void)
{
    // input mesh transformation
    string transformationFilename = MyConfigParser.Get("GenerateOneArffInput", "transformation_filename");
    R3Affine transformation = R3identity_affine;
    if (transformationFilename != "NULL") {
        transformation = FileUtil::ReadAffine(transformationFilename.c_str());
    }
    // input point cloud
    string pointCloudFilename = MyConfigParser.Get("GenerateOneArffInput", "point_cloud_filename");
    PointCloud pointCloud;
    pointCloud.ReadFile(pointCloudFilename.c_str());
    pointCloud.SetName(FileSystemUtil::GetBasename(pointCloudFilename, false).c_str());
    pointCloud.Transform(transformation);
    
    // input mesh
    string meshFilename = MyConfigParser.Get("GenerateOneArffInput", "mesh_filename");
    R3Mesh mesh;
    if (meshFilename != "NULL") {
        mesh.ReadFile(meshFilename.c_str());
        mesh.SetName(FileSystemUtil::GetBasename(meshFilename, false).c_str());
        mesh.Transform(transformation);
    }
    vector<string> descriptorNameArray;
    // descriptor name or descriptor name array
    if (!MyConfigParser.IsEnabled("GenerateOneArffInput", "process_descriptor_name_array_flag")) {
        // descriptor name
        descriptorNameArray = MyConfigParser.GetStringArray("GenerateOneArffInput", "descriptor_name");
    }
    else {
        descriptorNameArray = MyConfigParser.GetStringArray("GenerateOneArffInput", "descriptor_name_array");
    }
    vector<PointCloudPropertySet*> propertySetArray;
    if (MyConfigParser.IsEnabled("ShapeDescriptors", "generate_in_segments_separately")) {
        PointCloudPropertySet segmentation(&pointCloud);
        string fineGrainedSegmentationFilename = MyConfigParser.Get("GenerateOneArffInput", "segmentation_filename");
        segmentation.ReadFile(fineGrainedSegmentationFilename.c_str());
        propertySetArray = ShapeDescriptors::ComputeShapeDescriptorPerSegment(&mesh, &pointCloud, &segmentation, descriptorNameArray);
    }
    else {
        propertySetArray = ShapeDescriptors::ComputeShapeDescriptor(&mesh, &pointCloud, descriptorNameArray);
    }
    if (!MyConfigParser.IsEnabled("GenerateOneArffInput", "process_descriptor_name_array_flag")) {
        assert(false);
        if (propertySetArray.size() != 1) {
            fprintf(stderr, "GenerateOneArff generate %d properties from descriptorName = %s!\n",
                    propertySetArray.size(), descriptorNameArray[0].c_str());
            exit(-1);
        }
        // write to descriptors
        string outputDescriptorFilename = MyConfigParser.Get("GenerateOneArffOutput", "property_filename");
        propertySetArray[0]->WriteFile(outputDescriptorFilename.c_str());
    }
    else {
        string outputDescriptorDirname = MyConfigParser.Get("GenerateOneArffOutput", "property_dirname");
        for (int i=0; i<(int)propertySetArray.size(); i++) {
        string outputDescriptorFilename = outputDescriptorDirname + "/" + descriptorNameArray[i] + ".arff";
        propertySetArray[i]->WriteFile(outputDescriptorFilename.c_str());
        }
    }
    return 1;
}

static int
GenerateRangePerArffs(void)
{
    // input point cloud
    string pointCloudDirname = MyConfigParser.Get("GenerateRangePerArffsInput", "point_cloud_dirname");
    string propertyDirname = MyConfigParser.Get("GenerateRangePerArffsInput", "property_dirname");
    string segmentationDirname = MyConfigParser.Get("GenerateRangePerArffsInput", "segmentation_dirname");
    vector<string> propertyFilenameArray = FileSystemUtil::ListFiles(propertyDirname, "arff");
    vector<PointCloudPropertySet*> propertyArray;
    vector<PointCloudPropertySet*> segmentationArray;
    for (int i=0; i<(int)propertyFilenameArray.size(); i++) {
        string propertyFilename = propertyFilenameArray[i];
        string basename = FileSystemUtil::GetBasename(propertyFilename, false);
        string pointCloudFilename = pointCloudDirname+"/"+basename+".pts";
        string segmentationFilename = segmentationDirname+"/"+basename+".arff";
        if (!FileSystemUtil::PathExist(propertyFilename)) {
            fprintf(stderr, "GenerateRangePerArffs unable to find %s!\n",
                    propertyFilename.c_str());
            exit(-1);
        }
        if (!FileSystemUtil::PathExist(segmentationFilename)) {
            fprintf(stderr, "GenerateRangePerArfss unable to find %s!\n",
                    segmentationFilename.c_str());
            exit(-1);
        }
        // point cloud
        PointCloud* pCloud = new PointCloud;
        pCloud->ReadFile(pointCloudFilename.c_str());
        // segmentation
        PointCloudPropertySet* pSegmentation = new PointCloudPropertySet(pCloud);
        pSegmentation->ReadFile(segmentationFilename.c_str());
        segmentationArray.push_back(pSegmentation);
        // property
        PointCloudPropertySet* pProperty = new PointCloudPropertySet(pCloud);
        pProperty->ReadFile(propertyFilename.c_str());
        propertyArray.push_back(pProperty);
    }
    if (propertyArray.size() != segmentationArray.size()) {
        fprintf(stderr, "GenerateRangePerArffs size of propertyArray and segmentationArray doesnt match!\n");
        exit(-1);
    }
    if (propertyArray.size() == 0) {
        fprintf(stderr, "GenerateRangePerArffs propertyArray.size() == 0\n");
        exit(-1);
    }
    // get all segment names and all property names
    set<string> segmentNameArray;
    set<string> propertyNameArray;
    for (int meshIdx=0; meshIdx<(int)segmentationArray.size(); meshIdx++) {
        for (int segIdx=0; segIdx<segmentationArray[meshIdx]->NProperties(); segIdx++) {
            segmentNameArray.insert(segmentationArray[meshIdx]->Property(segIdx)->Name());
        }
    }
    for (int meshIdx=0; meshIdx<(int)propertyArray.size(); meshIdx++) {
        for (int prpIdx=0; prpIdx<propertyArray[meshIdx]->NProperties(); prpIdx++) {
            propertyNameArray.insert(propertyArray[meshIdx]->Property(prpIdx)->Name());
        }
    }
    printf("Segment names: \n    ");
    for (set<string>::const_iterator it=segmentNameArray.begin(); it!=segmentNameArray.end(); it++) {
        printf("%s, ",(*it).c_str());
    }
    printf("\n");
    printf("Property names: \n    ");
    for (set<string>::const_iterator it=propertyNameArray.begin(); it!=propertyNameArray.end(); it++) {
        printf("%s, ", (*it).c_str());
    }
    printf("\n");
    // mapping from segment name+property name to value array
    map<string, vector<double> > valueArrayMapping;
    // per mesh
    for (int meshIdx=0; meshIdx<(int)segmentationArray.size(); meshIdx++) {
        // per segment
        for (set<string>::const_iterator segIt=segmentNameArray.begin();
             segIt!=segmentNameArray.end(); segIt++) {
            PointCloudProperty* pSeg = segmentationArray[meshIdx]->Property(*segIt);
            if (!pSeg) continue;
            // per property
            for (set<string>::const_iterator prpIt=propertyNameArray.begin();
                 prpIt!=propertyNameArray.end(); prpIt++) {
                PointCloudProperty* pPrp = propertyArray[meshIdx]->Property(*prpIt);
                if (!pPrp) continue;
                // push back values
                string keyword = *segIt + ":" + *prpIt;
                // assert
                if (pSeg->NVertexValues() != pPrp->NVertexValues()) {
                    fprintf(stderr, "GenerateRangePerArffs pSeg and pPrp dont match!\n");
                    exit(-1);
                }
                // existencen of keyword
                if (valueArrayMapping.find(keyword) == valueArrayMapping.end()) {
                    valueArrayMapping[keyword] = vector<double>();
                }
                // push values
                for (int vtxIdx=0; vtxIdx<pPrp->NVertexValues(); vtxIdx++) {
                    if (pSeg->VertexValue(vtxIdx) < 0.5) continue;
                    valueArrayMapping[keyword].push_back(pPrp->VertexValue(vtxIdx));
                }
            }
        }
    }
    printf("Value array information ...\n");
    double minPercentile = MyConfigParser.GetScalar("Histogram", "min_cutoff_percentile");
    double maxPercentile = MyConfigParser.GetScalar("Histogram", "max_cutoff_percentile");
    for (map<string, vector<double> >::const_iterator it=valueArrayMapping.begin(); it!=valueArrayMapping.end(); it++) {
        const string& keyword = it->first;
        const vector<double>& valueArray = it->second;
        printf("    %s[%d values]: %g%% = %g, %g%% = %g\n",
               keyword.c_str(), (int)valueArray.size(),
               minPercentile*100.0,
               MathUtil::PercentileMedian(valueArray, minPercentile),
               maxPercentile*100.0,
               MathUtil::PercentileMedian(valueArray, maxPercentile));
    }
    string outDescriptionFilename = MyConfigParser.Get("GenerateRangePerArffsOutput", "range_description_filename");
    ofstream fout(outDescriptionFilename.c_str());
    for (map<string, vector<double> >::const_iterator it=valueArrayMapping.begin(); it!=valueArrayMapping.end(); it++) {
        const string& keyword = it->first;
        const vector<double>& valueArray = it->second;
        fout<<keyword<<" "<<MathUtil::PercentileMedian(valueArray, minPercentile)<<" "
        <<MathUtil::PercentileMedian(valueArray, maxPercentile)<<endl;
    }
    fout.close();
    return 1;
}

static void
ParseRangeDescriptionFile(const string& filename,
                          map<string, double>& minMapping, map<string, double>& maxMapping)
{
    ifstream fin(filename.c_str());
    string keyword;
    double minVal, maxVal;
    while (fin>>keyword>>minVal>>maxVal) {
        minMapping[keyword] = minVal;
        maxMapping[keyword] = maxVal;
    }
    fin.close();
}

static void
WriteHistogramFile(const string& filename,
                   const vector<double>& histogram)
{
    ofstream fout(filename.c_str());
    for (int i=0; i<(int)histogram.size(); i++) {
        fout<<histogram[i]<<" ";
    }
    fout.close();
}

static int
GenerateHistogram(void)
{
    map<string, double> minMapping, maxMapping;
    ParseRangeDescriptionFile(MyConfigParser.Get("GenerateHistogramInput", "range_description_filename"), minMapping, maxMapping);
    // input point cloud
    string pointCloudFilename = MyConfigParser.Get("GenerateHistogramInput", "point_cloud_filename");
    PointCloud pointCloud;
    pointCloud.ReadFile(pointCloudFilename.c_str());
    pointCloud.SetName(FileSystemUtil::GetBasename(pointCloudFilename, false).c_str());
    // input segmentation
    string segmentationFilename = MyConfigParser.Get("GenerateHistogramInput", "segmentation_filename");
    PointCloudPropertySet segmentation(&pointCloud);
    segmentation.ReadFile(segmentationFilename.c_str());
    // input property set
    string propertyFilename = MyConfigParser.Get("GenerateHistogramInput", "property_filename");
    PointCloudPropertySet propertySet(&pointCloud);
    propertySet.ReadFile(propertyFilename.c_str());
    if (propertySet.NProperties() != 1) {
        fprintf(stderr, "GenerateHistogram can only process a single property!\n");
        exit(-1);
    }
    // output histogram dirname
    string histogramDirname = MyConfigParser.Get("GenerateHistogramOutput", "histogram_dirname");
    // go through all segments
    int numBins = MyConfigParser.GetInt("Histogram", "number_of_bins_per_histogram");
    vector<string> formatArray = MyConfigParser.GetStringArray("Histogram", "format_array");
    for (int segIdx=0; segIdx<segmentation.NProperties(); segIdx++) {
        PointCloudProperty* pSeg = segmentation.Property(segIdx);
        // the first property
        PointCloudProperty* pPrp = propertySet.Property(0);
        string keyword = pSeg->Name()+":"+pPrp->Name();
        if (minMapping.find(keyword) == minMapping.end()) {
            fprintf(stderr, "GenerateHistogram unable to find %s!\n",
                    keyword.c_str());
            exit(-1);
        }
        double minVal = minMapping[keyword];
        if (maxMapping.find(keyword) == maxMapping.end()) {
            fprintf(stderr, "GenerateHistogram unable to find %s!\n",
                    keyword.c_str());
            exit(-1);
        }
        double maxVal = maxMapping[keyword];
        vector<double> histogram;
        vector<double> originalHist = pPrp->HistogramGivenMask(numBins, minVal, maxVal, pSeg);
        for (int formatIdx=0; formatIdx<(int)formatArray.size(); formatIdx++) {
            if (formatArray[formatIdx] == "original") {
                histogram.insert(histogram.end(), originalHist.begin(), originalHist.end());
            }
            else if (formatArray[formatIdx] == "cumulative") {
                vector<double> cumulativeHist = MathUtil::GenerateCumulativeHistogram(originalHist);
                cumulativeHist.pop_back();
                histogram.insert(histogram.end(), cumulativeHist.begin(), cumulativeHist.end());
            }
            else if (formatArray[formatIdx] == "mean_and_variance") {
                double mean = pPrp->MeanGivenMask(pSeg, minVal, maxVal);
                double variance = pPrp->StandardDeviationGivenMask(pSeg, minVal, maxVal);
                histogram.push_back(mean);
                histogram.push_back(variance);
            }
        }
        // only need segmentation name
        string histogramFilename = histogramDirname + "/" + pSeg->Name() + ".sd";
        WriteHistogramFile(histogramFilename, histogram);
    }
    return 1;
}

static int
Dispatcher(void) {
    string taskName = MyConfigParser.Get("Task", "task");
    if (taskName == "GenerateOneArff") {
        return GenerateOneArff();
    }
    else if (taskName == "GenerateRangePerArffs") {
        return GenerateRangePerArffs();
    }
    else if (taskName == "GenerateHistogram") {
        return GenerateHistogram();
    }
    else {
        fprintf(stderr, "Dispatcher unable to process %s!\n", taskName.c_str());
        exit(-1);
        return 0;
    }
}

int main(int argc, char **argv)
{
	// Configs
    MyConfigParser.Import(common_setting_filename, TRUE);
    MyConfigParser.Import(specific_setting_filename, TRUE);
	if (!ParseArgs(argc, argv)) exit(-1);
    Dispatcher();
	return 0;
}
