#ifndef MY_POINT_CLOUD_VIEWER_H
#define MY_POINT_CLOUD_VIEWER_H
// include glut related headers
#include "Utils/Utils.h"
#include <string>
#include "R3MyGraphics/R3MyGraphics.h"
namespace PointCloudViewer {
	// Functions
	void Init(int* argc, char **argv, const std::string& windowName, int width, int height,
              PointCloud* pCloud, PointCloudPropertySet* pPropertySet,
              Voxelization* pVoxel, std::vector<R3Box>* pCuboidArray);
	void ConfigInit(void);
	void LightingInit(void);
    void ReadRegionArray(const std::string& filename);
	R3Viewer * CreateBirdsEyeViewer(void);
    void TakeSnapshot(const std::string& filename);
    double GetNormalizedValue(PointCloudVertex* pVertex);
    RNRgb GetNormalizedColor(PointCloudVertex* pVertex);
    PointCloudProperty* GetCurrentProperty(void);
    void Draw(void);
	// Variables
    /*Window size*/
	extern int window_width;
	extern int window_height;
	/*Camera related*/
	extern R3Vector initial_camera_towards;
	extern R3Vector initial_camera_up;
	extern R3Point initial_camera_origin;
	extern RNBoolean initial_camera_flag;
	extern R3Viewer* viewer;
	extern bool exit_flag;
    /*Snapshot related*/
    extern std::string saved_image_name;
    extern int saved_image_counter;
    /*Data*/
    extern PointCloud* m_pPointCloud;
    extern PointCloudPropertySet* m_pPropertySet;
    extern Voxelization* m_pVoxelization;
    extern std::vector<R3Box>* m_pCuboidArray;
    extern int m_iCurrentPropertyIndex;
    extern std::vector<std::vector<int> > region_array;
    extern int region_center_node_index;
    extern RNInterval value_range;
    extern RNInterval percentile_range;
    /*Flags*/
    extern bool show_values_in_text_flag;
    extern bool show_statistics_flag;
    extern bool show_normals_flag;
    extern bool show_node_indices_flag;
    extern bool show_selected_region_flag;
    extern bool show_voxel_flag;
    extern bool show_cuboids_flag;
    extern bool show_points_flag;
    
#ifdef GLUT_RENDERING
    void GLUTInit(int* argc, char **argv, const std::string& windowName, int width, int height);
	void GLUTMainLoop(void);
	void GLUTIdle(void);
	void GLUTStop(void);
	void GLUTDraw(void);
	void GLUTResize(int w, int h);
	void GLUTKeyboard(unsigned char key, int x, int y);
	void GLUTSpecial(int key, int x, int y);
	void GLUTMouse(int button, int state, int x, int y);
	void GLUTMotion(int x, int y);
	void SafeStop(void);
	void FirstIdle(void);
    void PrintCameraSetting(void);
	void PrintHelper(void);
	extern int GLUTwindow;
	extern int GLUTmouse[2];
	extern int GLUTbutton[3];
	extern int GLUTmouse_drag;
	extern int GLUTmodifiers;
#elif defined MESA_RENDERING
    void MesaInit(int width, int height);
    void MesaDraw(void);
    void MesaExit(void);
    extern OSMesaContext m_MesaContext;
    extern unsigned char * m_aOffscreenBuffer;
#endif
}
#endif
