#ifdef GLUT_RENDERING
#include "PointCloudViewer.h"
#include "ConfigParser/ConfigParser.h"
int PointCloudViewer::GLUTwindow = 0;
int PointCloudViewer::GLUTmouse[2] = {0, 0};
int PointCloudViewer::GLUTbutton[3] = {0, 0, 0};
int PointCloudViewer::GLUTmouse_drag = 0;
int PointCloudViewer::GLUTmodifiers = 0;
using namespace std;

void PointCloudViewer::
GLUTInit(int* argc, char **argv, const std::string& windowName, int width, int height)
{
	// Create window
	glutInit(argc, argv);
    glutInitWindowPosition(100, 100);
	glutInitWindowSize(width, height);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STENCIL);
    GLUTwindow = glutCreateWindow(windowName.c_str());
    // Initialize GLUT callback functions
    glutDisplayFunc(GLUTDraw);
    glutIdleFunc(GLUTIdle);
    glutReshapeFunc(GLUTResize);
    glutKeyboardFunc(GLUTKeyboard);
    glutSpecialFunc(GLUTSpecial);
    glutMouseFunc(GLUTMouse);
    glutMotionFunc(GLUTMotion);
}

void PointCloudViewer::
GLUTMainLoop(void)
{
	// Run main loop -- never returns
    glutMainLoop();
}

void PointCloudViewer::
SafeStop(void)
{
    // don't support snapshot_all_dims_exit_flag
    if (MyConfigParser.IsEnabled("ExitConfig", "snapshot_on_exit_flag")) {
        TakeSnapshot(MyConfigParser.Get("ExitConfig", "snapshot_filename"));
    }
	exit_flag = 1;
	glutPostRedisplay();
}

void PointCloudViewer::
GLUTStop(void)
{
    // Destroy window
    glutDestroyWindow(GLUTwindow);
    // Exit
    exit(0);
}

void PointCloudViewer::
FirstIdle(void)
{
    if (MyConfigParser.IsEnabled("ExitConfig", "exit_immediately"))
        SafeStop();
	glutPostRedisplay();
}

void PointCloudViewer::
GLUTIdle(void)
{
	static int idle_redraw_counter = 0;
	static RNBoolean first_idle_flag = TRUE;
	if (idle_redraw_counter < 5)
    {
        idle_redraw_counter ++;
        glutPostRedisplay();
    }
    else if (first_idle_flag)
    {
        first_idle_flag = FALSE;
		FirstIdle();
    }
	else if (exit_flag) GLUTStop();
}

void PointCloudViewer::
GLUTDraw(void)
{
    // General draw
    Draw();
    // GLUT draw
    if (show_values_in_text_flag) {
        char buffer[1024];
        glDisable(GL_LIGHTING);
        RNLength d = 0.02 * m_pPointCloud->BBox().DiagonalRadius();
        for (int i = 0; i < m_pPointCloud->NPoints(); i++) {
            PointCloudVertex* pVertex = m_pPointCloud->Vertex(i);
            const R3Point& position = pVertex->Position();
            const R3Vector& normal = pVertex->Normal();
            sprintf(buffer, "%g", GetCurrentProperty()->VertexValue(pVertex));
            GLUTUtil::DrawText(position + d * normal, buffer);
        }
    }
    // draw node indices
    if (show_node_indices_flag) {
        char buffer[1024];
        glDisable(GL_LIGHTING);
        RNLength d = 0.02 * m_pPointCloud->BBox().DiagonalRadius();
        for (int i = 0; i < m_pPointCloud->NPoints(); i++) {
            PointCloudVertex* pVertex = m_pPointCloud->Vertex(i);
            const R3Point& position = pVertex->Position();
            const R3Vector& normal = pVertex->Normal();
            sprintf(buffer, "%d", i);
            GLUTUtil::DrawText(position + d * normal, buffer);
        }
    }
    // draw statistics
    if (show_statistics_flag && GetCurrentProperty()) {
        double y = window_height - 20;
        GLUTUtil::DrawText(R2Point(10, y), GetCurrentProperty()->Name().c_str(), viewer); y -= 20;
        string buffer = string("Mean:")+StringUtil::Number2String(GetCurrentProperty()->Mean());
        GLUTUtil::DrawText(R2Point(10, y), buffer.c_str(), viewer); y-= 20;
        buffer = string("Min:")+StringUtil::Number2String(GetCurrentProperty()->Minimum());
        GLUTUtil::DrawText(R2Point(10, y), buffer.c_str(), viewer); y-= 20;
        buffer = string("Max:")+StringUtil::Number2String(GetCurrentProperty()->Maximum());
        GLUTUtil::DrawText(R2Point(10, y), buffer.c_str(), viewer); y-= 20;
        buffer = string("Stddev:")+StringUtil::Number2String(GetCurrentProperty()->StandardDeviation());
        GLUTUtil::DrawText(R2Point(10, y), buffer.c_str(), viewer); y-= 20;
        buffer = string("Min Display Value:")+StringUtil::Number2String(value_range.Min());
        GLUTUtil::DrawText(R2Point(10, y), buffer.c_str(), viewer); y-= 20;
        buffer = string("Max Display Value:")+StringUtil::Number2String(value_range.Max());
        GLUTUtil::DrawText(R2Point(10, y), buffer.c_str(), viewer); y-= 20;
        buffer = string("Min Display Percentile:")+StringUtil::Number2String(percentile_range.Min());
        GLUTUtil::DrawText(R2Point(10, y), buffer.c_str(), viewer); y-= 20;
        buffer = string("Max Display Percentile:")+StringUtil::Number2String(percentile_range.Max());
        GLUTUtil::DrawText(R2Point(10, y), buffer.c_str(), viewer); y-= 20;
    }
    // take snapshot if necessary
    if (saved_image_name != "") {
        R2Image image(window_width, window_height, 3);
        image.Capture();
        image.Write(saved_image_name.c_str());
        printf("PointCloudViewer::GLUTDraw Take a snapshot and write to image %s ...\n", saved_image_name.c_str());
        saved_image_name = "";
    }
    // Swap buffers
    glutSwapBuffers();
}

void PointCloudViewer::
GLUTResize(int w, int h)
{
	// Resize window
	glViewport(0, 0, w, h);
	// Resize viewer viewport
	viewer->ResizeViewport(0, 0, w, h);
	// Remember window size
	window_width = w;
	window_height = h;
	// Redraw
	glutPostRedisplay();
}

void PointCloudViewer::
GLUTKeyboard(unsigned char key, int x, int y)
{
	// Process keyboard button event
	switch (key) {
        case 'c':
            PrintCameraSetting();
            break;
        case 'n':
            show_normals_flag = !show_normals_flag;
            break;
        case 'T':
            show_node_indices_flag = !show_node_indices_flag;
            break;
        case 's':
            show_statistics_flag = !show_statistics_flag;
            break;
        case 't':
            show_values_in_text_flag = !show_values_in_text_flag;
            break;
        case 'r':
            show_selected_region_flag = !show_selected_region_flag;
            break;
        case 'R':
        {
            int node_index = -1;
            printf("Please input the index of the centering node of the next region ([%d,%d])?",
                   0,(int)region_array.size()-1);
            scanf("%d", &node_index);
            if (node_index >= (int)region_array.size() || node_index<0) {
                printf("Index is out of range!\n");
            }
            else {
                region_center_node_index = node_index;
            }
            break;
        }
        case 'H':
            PrintHelper();
            break;
        case 'G':
        {
            int node_index = -1;
            printf("Please input the index of the query node ([%d,%d])?",
                   0,m_pPointCloud->NPoints()-1);
            scanf("%d", &node_index);
            if (node_index >= m_pPointCloud->NPoints() || node_index<0) {
                printf("Index is out of range!\n");
            }
            else {
                R3Point position = m_pPointCloud->Vertex(node_index)->Position();
                printf("Position of node %d: [%g %g %g]\n", node_index,
                       position[0], position[1], position[2]);
            }
            break;
        }
        case 'v':
            show_voxel_flag = !show_voxel_flag;
            break;
        case 'C':
            show_cuboids_flag = !show_cuboids_flag;
            break;
        case 'p':
            show_points_flag = !show_points_flag;
            break;
        case '~':
            saved_image_name = string("image_cap_")+StringUtil::Number2String(saved_image_counter)+".png";
            saved_image_counter ++;
            break;
        case 27: // ESCAPE
			GLUTStop();
			break;
	}
	
	// Remember mouse position
	GLUTmouse[0] = x;
	GLUTmouse[1] = window_height - y;
	
	// Remember modifiers
	GLUTmodifiers = glutGetModifiers();
	
	// Redraw
	glutPostRedisplay();
}


void PointCloudViewer::
GLUTSpecial(int key, int x, int y)
{
	// Invert y coordinate
	y = window_height - y;
	
	// Process keyboard button event
    // Process keyboard button event
    switch (key) {
        case GLUT_KEY_HOME:
        case GLUT_KEY_END:
        case GLUT_KEY_PAGE_DOWN:
        case GLUT_KEY_PAGE_UP: {
            if (key == GLUT_KEY_PAGE_DOWN) m_iCurrentPropertyIndex--;
            else if (key == GLUT_KEY_PAGE_UP) m_iCurrentPropertyIndex++;
            else if (key == GLUT_KEY_HOME) m_iCurrentPropertyIndex = 0;
            else if (key == GLUT_KEY_END) m_iCurrentPropertyIndex = m_pPropertySet->NProperties() - 1;
            if (m_iCurrentPropertyIndex < 0)
                m_iCurrentPropertyIndex = 0;
            if (m_iCurrentPropertyIndex >= m_pPropertySet->NProperties())
                m_iCurrentPropertyIndex = m_pPropertySet->NProperties() - 1;
            // percentile_range.Reset(10, 90);
            RNScalar min_value = GetCurrentProperty()->Percentile(percentile_range.Min());
            RNScalar max_value = GetCurrentProperty()->Percentile(percentile_range.Max());
            value_range.Reset(min_value, max_value);
            break; }
            
        case GLUT_KEY_RIGHT:
        case GLUT_KEY_LEFT:
        case GLUT_KEY_DOWN:
        case GLUT_KEY_UP: {
            RNScalar mid = percentile_range.Mid();
            RNScalar radius = percentile_range.Radius();
            if (key == GLUT_KEY_LEFT) { if (radius >= 1) radius -= 1; }
            else if (key == GLUT_KEY_RIGHT) { if ((mid - radius >= 1) && (mid + radius <= 99)) radius += 1; }
            else if (key == GLUT_KEY_DOWN) { if (mid - radius >= 1) mid -= 1; }
            else if (key == GLUT_KEY_UP) { if (mid + radius <= 99) mid += 1; }
            percentile_range.Reset(mid - radius, mid + radius);
            RNScalar min_value = GetCurrentProperty()->Percentile(percentile_range.Min());
            RNScalar max_value = GetCurrentProperty()->Percentile(percentile_range.Max());
            value_range.Reset(min_value, max_value);
            break; }
    }
    
	// Remember mouse position
	GLUTmouse[0] = x;
	GLUTmouse[1] = y;
	
	// Remember modifiers
	GLUTmodifiers = glutGetModifiers();
	
	// Redraw
	glutPostRedisplay();
}

void PointCloudViewer::
GLUTMouse(int button, int state, int x, int y)
{
	// Invert y coordinate
	y = window_height - y;
	
	// Mouse is going down
	if (state == GLUT_DOWN) {
		// Reset mouse drag
		GLUTmouse_drag = 0;
		// left
		if (button == GLUT_LEFT_BUTTON) {
			// Todo
		}
	}
	else {
	}
	
	// Remember button state
	int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
	GLUTbutton[b] = (state == GLUT_DOWN) ? 1 : 0;
	
	// Remember modifiers
	GLUTmodifiers = glutGetModifiers();
	
	// Remember mouse position
	GLUTmouse[0] = x;
	GLUTmouse[1] = y;
	
	// Redraw
	glutPostRedisplay();
}

void PointCloudViewer::
GLUTMotion(int x, int y)
{
	// Invert y coordinate
	y = window_height - y;
	
	// Compute mouse movement
	int dx = x - GLUTmouse[0];
	int dy = y - GLUTmouse[1];
	
	// Update mouse drag
	GLUTmouse_drag += dx*dx + dy*dy;
	
	// Drag object
	if (GLUTbutton[0] && FALSE) {
		// Todo
	}
	// Drag viewer
	R3Point rotation_center = m_pPointCloud->Centroid();
	if (GLUTbutton[0]) {
		viewer->RotateWorld(1.0, rotation_center, x, y, dx, dy);
	}
	else if (GLUTbutton[1]) {
		viewer->ScaleWorld(1.0, rotation_center, x, y, dx, dy);
	}
	else if (GLUTbutton[2]) {
		viewer->TranslateWorld(1.0, rotation_center, x, y, dx, dy);
	}
	if (GLUTbutton[0] || GLUTbutton[1] || GLUTbutton[2]) {
		glutPostRedisplay();
	}
	
	// Remember mouse position
	GLUTmouse[0] = x;
	GLUTmouse[1] = y;
}

void PointCloudViewer::
PrintHelper(void)
{
    printf("PointCloudViewer::PrintHelper\n");
    printf("    c    :    print camera parameters\n");
    printf("    n    :    turn on/off normals\n");
    printf("    s    :    turn on/off statistics\n");
    printf("    t    :    turn on/off values\n");
    printf("    T    :    turn on/off node indices\n");
    printf("    r    :    turn on/off selected region (\'selected_region_filename\')\n");
    printf("    v    :    turn on/off voxels\n");
    printf("    C    :    turn on/off cuboids\n");
    printf("    p    :    turn on/off points\n");
    printf("    R    :    change the region index\n");
    printf("    G    :    print position of a point\n");
    printf("\n");
}

void PointCloudViewer::
PrintCameraSetting(void)
{
    const R3Camera& camera = viewer->Camera();
    const R3Vector& up = camera.Up();
    const R3Vector& towards = camera.Towards();
    printf("PointCloudViewer::PrintCameraSetting\n    towards = %g,%g,%g\n    up      = %g,%g,%g\n",
           towards[0], towards[1], towards[2],
           up[0], up[1], up[2]);
}
#endif


