#ifdef MESA_RENDERING
#include "ConfigParser/ConfigParser.h"
#include "PointCloudViewer.h"
using namespace std;
OSMesaContext PointCloudViewer::m_MesaContext;
unsigned char * PointCloudViewer::m_aOffscreenBuffer;
void PointCloudViewer::
MesaInit(int width, int height)
{
#if OSMESA_MAJOR_VERSION * 100 + OSMESA_MINOR_VERSION >= 305
	/* specify Z, stencil, accum sizes */
	m_MesaContext = OSMesaCreateContextExt( OSMESA_RGBA, 16, 0, 0, NULL );
#else
	m_MesaContext = OSMesaCreateContext( OSMESA_RGBA, NULL );
#endif
    if (!m_MesaContext) {
        fprintf(stderr, "PointCloudViewer::MesaInit OSMesaCreateContext failed!\n");
        exit(-1);
    }
    m_aOffscreenBuffer = new unsigned char[4*width*height];
    bool ok = OSMesaMakeCurrent(m_MesaContext, m_aOffscreenBuffer, GL_UNSIGNED_BYTE, width, height);
    if (!ok) {
        fprintf(stderr, "PointCloudViewer::MesaInit OSMesaMakeCurrent failed!\n");
        exit(-1);
    }
}

void PointCloudViewer::
MesaDraw(void)
{
    if (MyConfigParser.IsEnabled("ExitConfig", "snapshot_on_exit_flag")) {
        Draw();
        TakeSnapshot(MyConfigParser.Get("ExitConfig", "snapshot_filename"));
        R2Image image(window_width, window_height, 4, m_aOffscreenBuffer);
        image.Write(saved_image_name.c_str());
        printf("PointCloudViewer::MesaDraw Take a snapshot and write to image %s ...\n", saved_image_name.c_str());
        saved_image_name = "";
    }
    else if (MyConfigParser.IsEnabled("ExitConfig", "snapshot_all_dims_on_exit_flag")) {
        string dirname = MyConfigParser.Get("ExitConfig", "snapshot_all_dims_dirname");
        FileSystemUtil::CreateDirectory(dirname);
        for (int i=0; i<m_pPropertySet->NProperties(); i++) {
            m_iCurrentPropertyIndex = i;
            //RNScalar min_value = GetCurrentProperty()->Percentile(percentile_range.Min());
            //RNScalar max_value = GetCurrentProperty()->Percentile(percentile_range.Max());
            //value_range.Reset(min_value, max_value);
            string propertyName = GetCurrentProperty()->Name();
            Draw();
            TakeSnapshot(dirname+"/"+propertyName+".png");
            R2Image image(window_width, window_height, 4, m_aOffscreenBuffer);
            image.Write(saved_image_name.c_str());
            printf("PointCloudViewer::MesaDraw Take a snapshot and write to image %s ...\n", saved_image_name.c_str());
            saved_image_name = "";
        }
    }
}

void PointCloudViewer::
MesaExit(void)
{
    OSMesaDestroyContext(m_MesaContext);
    delete[] m_aOffscreenBuffer;
    m_aOffscreenBuffer = NULL;
}
#endif
