#include <vector>
#include <fstream>
#include "R3MyGraphics/R3MyGraphics.h"
#include "ConfigParser/ConfigParser.h"
#include "PointCloudViewer.h"
#include "Utils/Utils.h"
using namespace std;
static char common_setting_filename[] = "./configs/config.cfg";
static char specific_setting_filename[] = "./configs/point_cloud_viewer.cfg";
static vector<string> arg_list_secs;
static vector<string> arg_list_opts;
static vector<string> arg_list_vals;
static string included_config_filename = "NULL";

// we are able to import an external config file by using arguments
static int
ParseArgs(int argc, char **argv)
{
    // Parse arguments
    argc--; argv++;
    while (argc > 0) {
        if ((*argv)[0] == '-') {
			if (!strcmp(*argv, "-Config"))
			{
				argv++; argc--;
                included_config_filename = string(*argv);
			}
            else if (!strcmp(*argv, "-Help")) {
                MyConfigParser.PrintAllSettings();
                exit(-1);
            }
			else {
                char* sec = (*argv)+1;
                argv++; argc--;
                char* opt = *argv;
                argv++; argc--;
                char* val = *argv;
                arg_list_secs.push_back(sec);
                arg_list_opts.push_back(opt);
                arg_list_vals.push_back(val);
			}
            argv++; argc--;
        }
        else {
            fprintf(stderr, "Invalid program argument: %s", *argv); exit(1);
            argv++; argc--;
        }
    }
    MyConfigParser.ImportFrequentlyUsedParameters();
    // Return OK status
    return 1;
}

static int
ExecuteArgs(void)
{
    for (int i=0; i<(int)arg_list_secs.size(); i++) {
        string sec = arg_list_secs[i];
        string opt = arg_list_opts[i];
        string val = arg_list_vals[i];
        if (MyConfigParser.HasSection(sec)) {
            MyConfigParser.Set(sec, opt, val);
        }
        else {
            fprintf(stderr, "Invalid program argument: %s", sec.c_str());
            exit(-1);
        }
    }
    MyConfigParser.ImportFrequentlyUsedParameters();
    return 1;
}


int main(int argc, char** argv)
{
	// Configs
    MyConfigParser.Import(common_setting_filename, TRUE);
	MyConfigParser.Import(specific_setting_filename, TRUE);
    // parse args: check if there is -config
    if (!ParseArgs(argc, argv)) exit(-1);
	if (!ExecuteArgs()) exit(-1);
    // point cloud
    PointCloud pointCloud;
    pointCloud.ReadFile(MyConfigParser.Get("Input", "point_cloud_filename").c_str());
    // point cloud property
    PointCloudPropertySet propertySet(&pointCloud);
    string propertySetFilename = MyConfigParser.Get("Input", "property_set_filename");
    if (propertySetFilename != "NULL") {
        propertySet.ReadFile(propertySetFilename.c_str());
    }
    // voxelization
    Voxelization voxel;
    string voxelizationFilename = MyConfigParser.Get("Input", "voxelization_filename");
    if (voxelizationFilename != "NULL") {
        voxel.ReadFile(voxelizationFilename.c_str());
    }
    // cuboids
    string cuboidsFilename = MyConfigParser.Get("Input", "cuboid_list_filename");
    vector<R3Box> cuboidArray;
    if (cuboidsFilename != "NULL") {
        FileUtil::ImportCuboids(cuboidArray, cuboidsFilename);
    }
    PointCloudViewer::Init(&argc, argv, "PointCloudViewer", 1000, 800, &pointCloud,
                           &propertySet, &voxel, &cuboidArray);
#ifdef GLUT_RENDERING
    PointCloudViewer::GLUTMainLoop();
#elif defined MESA_RENDERING
    PointCloudViewer::MesaDraw();
    PointCloudViewer::MesaExit();
#endif
    return 0;
}










