#include "Descriptor/ShapeDescriptors.h"
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
#include <vector>
using namespace std;
using namespace ShapeDescriptors;

// Some simple descriptors
const vector<double> ShapeDescriptors::
BBoxDimensions(PointCloud* pCloud)
{
    R3Box box = pCloud->BBox();
    vector<double> descs;
    descs.push_back(box.XLength());
    descs.push_back(box.YLength());
    descs.push_back(box.ZLength());
    MathUtil::DivideBySum(descs);
    return descs;
}

const vector<DescriptorsPerLabel> ShapeDescriptors::
BBoxDimensionsPerLabel(PointCloud* pCloud)
{
    vector<DescriptorsPerLabel> descArray;
    // all
    DescriptorsPerLabel desc;
    desc.labelName = "all";
    R3Box box = pCloud->BBox();
    desc.descriptors.push_back(box.XLength());
    desc.descriptors.push_back(box.YLength());
    desc.descriptors.push_back(box.ZLength());
    MathUtil::DivideBySum(desc.descriptors);
    descArray.push_back(desc);
    // per segment
    vector<R3Box> boxArray(pCloud->NSegments(), R3null_box);
    for (int i=0; i<pCloud->NPoints(); i++) {
        PointCloudVertex* pVtx = pCloud->Vertex(i);
        int segIndex = pVtx->SegmentIndex();
        boxArray[segIndex].Union(pVtx->Position());
    }
    for (int i=0; i<(int)boxArray.size(); i++) {
        DescriptorsPerLabel desc;
        desc.labelName = pCloud->KthSegmentName(i);
        desc.descriptors.push_back(boxArray[i].XLength());
        desc.descriptors.push_back(boxArray[i].YLength());
        desc.descriptors.push_back(boxArray[i].ZLength());
        MathUtil::DivideBySum(desc.descriptors);
        descArray.push_back(desc);
    }
    return descArray;
}