#include <string>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include "Descriptor/ShapeDescriptors.h"
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
#include "RNBasics/RNBasics.h"

using namespace std;
using namespace ShapeDescriptors;

LaplacianSpectrumExporter::
LaplacianSpectrumExporter(PointCloud* pCloud)
:GeodesicComputer(pCloud)
{}

// the paper "Mesh Saliency via Spectral Processing" is not so clear about this
// but what they actually did is
// 1. Compute geometric laplacian W (without normalization)
// 2. Each diagonal element in D is the sum of all off-diagonal elements in W
void LaplacianSpectrumExporter::
GenerateLaplacianMatrix(PointCloudProperty* mask, MathUtil::SparseMatrix& sparseMat)
{
    // connect indices
    vector<int> globalIndex2PartIndex(NSamples(), -1);
    vector<int> partIndex2GlobalIndex;
    for (int i=0; i<NSamples(); i++) {
        // only include samples in the segment
        if (mask->VertexValue(i) > 0.5) {
            globalIndex2PartIndex[i] = (int)partIndex2GlobalIndex.size();
            partIndex2GlobalIndex.push_back(i);
        }
    }
    // basics of sparse matrix
    sparseMat.m = partIndex2GlobalIndex.size();
    sparseMat.n = partIndex2GlobalIndex.size();
    // go through all indices in the part
    for (int partSrcIdx=0; partSrcIdx<(int)partIndex2GlobalIndex.size(); partSrcIdx++) {
        int globalSrcIdx = partIndex2GlobalIndex[partSrcIdx];
        // add diagonal elements in the end
        // off-diagonal
        vector<double> weights;
        for (int j=0; j<(int)m_vAdjacentIndexList[globalSrcIdx].size(); j++) {
            int globalTarIdx = m_vAdjacentIndexList[globalSrcIdx][j];
            int partTarIdx = globalIndex2PartIndex[globalTarIdx];
            if (partTarIdx == -1 || partTarIdx == partSrcIdx) continue;
            sparseMat.xs.push_back(partSrcIdx);
            sparseMat.ys.push_back(partTarIdx);
            RNLength distance = R3Distance(GetKthSample(globalSrcIdx).Position(), GetKthSample(globalTarIdx).Position());
            weights.push_back(1.0/distance/distance);
        }
        // don't normalization
        // MathUtil::DivideBySum(weights);
        // push back weights
        for (int j=0; j<(int)weights.size(); j++) {
            sparseMat.values.push_back(weights[j]);
        }
        // push back diagonal weights
        sparseMat.xs.push_back(partSrcIdx);
        sparseMat.ys.push_back(partSrcIdx);
        sparseMat.values.push_back(-MathUtil::Sum(weights));
    }
}

void LaplacianSpectrumExporter::
Run(void)
{
    printf("LaplacianSpectrumExporter::Run processing %d samples ...\n", NSamples());
    fprintf(stderr, "LaplacianSpectrumExporter::Run not updated yet!\n");
    exit(-1);
//    // construct connected graph and compute geodesic distance
//    ConstructConnectedGraphAndComputeGeodesicDistances();
//    // workspace
//    string workDirname = MyConfigParser.Get("LaplacianSpectrum", "intermediate_matrix_and_eigen_dirname");
//    FileSystemUtil::CreateDirectory(workDirname);
//    string eigenSolverScriptname = MyConfigParser.Get("LaplacianSpectrum",
//                                                      "eigen_solver_scriptname");
//    int resampledNumDims = MyConfigParser.GetInt("LaplacianSpectrum", "resampled_num_dimensions");
//    int clampedNumDims = MyConfigParser.GetInt("LaplacianSpectrum", "clamped_num_dimensions");
//    // create masks
//    PointCloudPropertySet maskSet(m_pPointCloud);
//    maskSet.CreateSegmentMasks();
//    for (int k=0; k<maskSet.NProperties(); k++) {
//        DescriptorsPerLabel descPerLabel;
//        descPerLabel.labelName = maskSet.Property(k)->Name();
//        // get sparse matrix
//        MathUtil::SparseMatrix sparseMat;
//        GenerateLaplacianMatrix(maskSet.Property(k), sparseMat);
//        // output to a file
//        string matrixFilename = workDirname + "/matrix-"+descPerLabel.labelName+".txt";
//        string eigenFilename = workDirname + "/eigen-"+descPerLabel.labelName+".txt";
//        vector<double> eigenValues =
//        sparseMat.SolveEigenvaluesUsingPython(eigenSolverScriptname,
//                                              matrixFilename, eigenFilename);
//        // convert all eigen values to positive numbers and sort ascendingly
//        for (int i=0; i<(int)eigenValues.size(); i++) eigenValues[i] = fabs(eigenValues[i]);
//        MathUtil::Sort(eigenValues, true);
//        // make sure the first eigenvalue is very close to 0
//        if (eigenValues[0] > 1e-6) {
//            fprintf(stderr, "[Warning]LaplacianSpectrumExporter::Run smallest eigvenvalue is %g > 1e-6!\n", eigenValues[0]);
//        }
//        // compute log
//        // skip the first eigenvalue
//        vector<double> logEigenvalues;
//        for (int i=1; i<(int)eigenValues.size(); i++) {
//            double val = eigenValues[i];
//            if (val < 1e-32) val = 1e-32;
//            logEigenvalues.push_back(log(val));
//        }
//        // subtract the median log value
//        double base = logEigenvalues[logEigenvalues.size()/2];
//        for (int i=0; i<(int)logEigenvalues.size(); i++) {
//            logEigenvalues[i] -= base;
//        }
//        vector<double> unclampedDescs = MathUtil::ResampleVector(logEigenvalues, resampledNumDims);
//        descPerLabel.descriptors = vector<double>(unclampedDescs.begin(), unclampedDescs.begin()+clampedNumDims);
//        m_vFeaturePerLabelArray.push_back(descPerLabel);
//    }
}

//const vector<double>& LaplacianSpectrumExporter::
//Get(void) const
//{
//    // find the feature array with label = all
//    for (int i=0; i<(int)m_vFeaturePerLabelArray.size(); i++) {
//        if (m_vFeaturePerLabelArray[i].labelName == "all") {
//            return m_vFeaturePerLabelArray[i].descriptors;
//        }
//    }
//    fprintf(stderr, "CurvatureHistogramExporter::Get unable to find 'all'!\n");
//    exit(-1);
//}
//
//const vector<DescriptorsPerLabel>& LaplacianSpectrumExporter::
//GetPerLabel(void) const
//{
//    return m_vFeaturePerLabelArray;
//}
//
//const vector<double> ShapeDescriptors::
//LaplacianSpectrum(PointCloud* pCloud)
//{
//    LaplacianSpectrumExporter exporter(pCloud);
//    exporter.Run();
//    return exporter.Get();
//}
//
//const vector<DescriptorsPerLabel> ShapeDescriptors::
//LaplacianSpectrumPerLabel(PointCloud* pCloud)
//{
//    LaplacianSpectrumExporter exporter(pCloud);
//    exporter.Run();
//    return exporter.GetPerLabel();
//}
