#include "Descriptor/ShapeDescriptors.h"
#include "ConfigParser/ConfigParser.h"
#include "Utils/Utils.h"
#include <vector>
using namespace std;
using namespace ShapeDescriptors;

const vector<double> ShapeDescriptors::
Material(const string& modelName)
{
    string materialSpecFilename = MyConfigParser.Get("ShapeDescriptors", "material_specification");
    if (!FileSystemUtil::PathExist(materialSpecFilename)) {
        fprintf(stderr, "ShapeDescriptors::Material unable to find material specification file %s!\n", materialSpecFilename.c_str());
        exit(-1);
    }
    // extract the first line of the file, which indicates the indices
    map<string, int> mappingMat2Index;
    vector<string> lines = FileUtil::ReadLines(materialSpecFilename);
    if ((int)lines.size()<1) {
        fprintf(stderr, "ShapeDescriptors::Material unable to find the first line of %s!\n", materialSpecFilename.c_str());
        exit(-1);
    }
    vector<string> segs = StringUtil::Split(lines[0], ',');
    for (int i=0; i<(int)segs.size(); i++) {
        mappingMat2Index[segs[i]] = i;
    }
    // initialize the descriptor
    vector<double> descriptors(mappingMat2Index.size(), 0.0);
    bool detectedFlag = false;
    // find the line that starts with the modelName, and read in the list
    for (int i=1; i<(int)lines.size(); i++) {
        vector<string> segs = StringUtil::Split(lines[i], ',');
        if (segs.size() < 1) continue;
        if (segs[0] != modelName) continue;
        if (detectedFlag) {
            fprintf(stderr, "ShapeDescriptors::Material there are two lines that begin with %s in %s!\n", modelName.c_str(), materialSpecFilename.c_str());
            exit(-1);
        }
        detectedFlag = true;
        for (int j=1; j<(int)segs.size(); j++) {
            string matName = segs[j];
            if (mappingMat2Index.find(matName) == mappingMat2Index.end()) {
                fprintf(stderr, "ShapeDescriptors::Material material %s is undefined (line %s in %s)!\n",
                        matName.c_str(), modelName.c_str(), materialSpecFilename.c_str());
                exit(-1);
            }
            int index = mappingMat2Index[matName];
            if (index >= (int)descriptors.size()) {
                fprintf(stderr, "ShapeDescriptors::Material index %d is out of range!\n", index);
                exit(-1);
            }
            descriptors[index] = 1.0;
        }
    }
    if (!detectedFlag) {
        fprintf(stderr, "ShapeDescriptors::Material unable to find a line for %s in %s!\n", modelName.c_str(), materialSpecFilename.c_str());
        exit(-1);
    }
    return descriptors;
}