#include <string>
#include "Descriptor/ShapeDescriptors.h"
#include "ConfigParser/ConfigParser.h"
#include "R3MyGraphics/R3MyGraphics.h"
using namespace std;
using namespace ShapeDescriptors;

//const vector<double> ShapeDescriptors::
//GetShapeDescriptors(R3Mesh* pMesh)
//{
//    PointCloud pointCloud;
//    int numSamples = MyConfigParser.GetInt("Sampling", "number_uniform_samples");
//    bool isPhongShading = MyConfigParser.IsEnabled("Sampling", "phong_shading_flag");
//    pointCloud.Import(pMesh, numSamples, isPhongShading);
//    if (MyConfigParser.Get("Sampling", "point_cloud_filename") != "NULL") {
//        pointCloud.WriteFile(MyConfigParser.Get("Sampling", "point_cloud_filename").c_str());
//    }
//    string type = MyConfigParser.Get("ShapeDescriptors", "type");
//    if (type == "material") {
//        // assuming pMesh->Name() gives the model name
//        return Material(pMesh->Name());
//    }
//    else if (type == "curvature_histogram") {
//        return CurvatureHistogram(&pointCloud);
//    }
//    else if (type == "shape_diameter_histogram") {
//        return ShapeDiameterHistogram(pMesh, &pointCloud);
//    }
//    else if (type == "laplacian_spectrum") {
//        return LaplacianSpectrum(&pointCloud);
//    }
//    else if (type == "d2") {
//        return D2(&pointCloud);
//    }
//    else if (type == "bbox_dimensions") {
//        return BBoxDimensions(&pointCloud);
//    }
//    else {
//        fprintf(stderr, "ShapeDescriptors::GetShapeDescriptors unable to "
//                "process type = %s!\n", type.c_str());
//        exit(-1);
//    }
//}
//
//const vector<double> ShapeDescriptors::
//GetShapeDescriptors(PointCloud* pCloud)
//{
//    string type = MyConfigParser.Get("ShapeDescriptors", "type");
//    if (type == "curvature_histogram") {
//        return CurvatureHistogram(pCloud);
//    }
//    else if (type == "laplacian_spectrum") {
//        return LaplacianSpectrum(pCloud);
//    }
//    else if (type == "d2") {
//        return D2(pCloud);
//    }
//    else if (type == "bbox_dimensions") {
//        return BBoxDimensions(pCloud);
//    }
//    else {
//        fprintf(stderr, "ShapeDescriptors::GetShapeDescriptors unable to "
//                "process type = %s!\n", type.c_str());
//        exit(-1);
//    }
//}
//
//const vector<DescriptorsPerLabel> ShapeDescriptors::
//GetShapeDescriptorsPerLabel(R3Mesh* pMesh)
//{
//    PointCloud pointCloud;
//    int numSamples = MyConfigParser.GetInt("Sampling", "number_uniform_samples");
//    bool isPhongShading = MyConfigParser.IsEnabled("Sampling", "phong_shading_flag");
//    pointCloud.Import(pMesh, numSamples, isPhongShading);
//    if (MyConfigParser.Get("Sampling", "point_cloud_filename") != "NULL") {
//        pointCloud.WriteFile(MyConfigParser.Get("Sampling", "point_cloud_filename").c_str());
//    }
//    string type = MyConfigParser.Get("ShapeDescriptors", "type");
//    if (type == "curvature_histogram") {
//        return CurvatureHistogramPerLabel(&pointCloud);
//    }
//    else if (type == "shape_diameter_histogram") {
//        return ShapeDiameterHistogramPerLabel(pMesh, &pointCloud);
//    }
//    else if (type == "laplacian_spectrum") {
//        return LaplacianSpectrumPerLabel(&pointCloud);
//    }
//    else if (type == "bbox_dimensions") {
//        return BBoxDimensionsPerLabel(&pointCloud);
//    }
//    else {
//        fprintf(stderr, "ShapeDescriptors::GetShapeDescriptorsPerLabel unable to "
//                "process type = %s!\n", type.c_str());
//        exit(-1);
//    }
//}

DescriptorExporter::
DescriptorExporter(R3Mesh* pMesh, PointCloud* pCloud)
: m_pMesh(pMesh), m_pPointCloud(pCloud), m_pPropertySet(NULL)
{}

PointCloudPropertySet* DescriptorExporter::
Get(void)
{
    Run();
    return m_pPropertySet;
}

void DescriptorExporter::
Run(void)
{}

vector<PointCloudPropertySet*> ShapeDescriptors::
ComputeShapeDescriptor(R3Mesh* pMesh, PointCloud* pCloud, const vector<string>& descriptorNameArray)
{
    ShapeDiameterExporter sdfExporter(pMesh, pCloud);
    CurvatureExporter curvatureExporter(pCloud);
    vector<PointCloudPropertySet*> propertySetArray;
    for (int i=0; i<(int)descriptorNameArray.size(); i++) {
        string descriptorName = descriptorNameArray[i];
        if (descriptorName == "ShapeDiameter") {
            propertySetArray.push_back(sdfExporter.Get());
        }
        else if (descriptorName == "CurvatureK1Large") {
            // set k1
            MyConfigParser.Set("CurvatureHistogram", "feature_array_per_scale", "K1OrderedAbs");
            // set a low normal_threshold
            MyConfigParser.Set("CurvatureHistogram", "normal_threshold", "-10000.0");
            propertySetArray.push_back(curvatureExporter.Get());
        }
        else if (descriptorName == "CurvatureK2Large") {
            // set k2
            MyConfigParser.Set("CurvatureHistogram", "feature_array_per_scale", "K2OrderedAbs");
            // set a low normal_threshold
            MyConfigParser.Set("CurvatureHistogram", "normal_threshold", "-10000.0");
            propertySetArray.push_back(curvatureExporter.Get());
        }
        else if (descriptorName == "CurvatureK1Small") {
            // set k1
            MyConfigParser.Set("CurvatureHistogram", "feature_array_per_scale", "K1OrderedAbs");
            // set a high normal_threshold
            MyConfigParser.Set("CurvatureHistogram", "normal_threshold", "0.5");
            propertySetArray.push_back(curvatureExporter.Get());
        }
        else if (descriptorName == "CurvatureK2Small") {
            // set k2
            MyConfigParser.Set("CurvatureHistogram", "feature_array_per_scale", "K2OrderedAbs");
            // set a high normal_threshold
            MyConfigParser.Set("CurvatureHistogram", "normal_threshold", "0.5");
            propertySetArray.push_back(curvatureExporter.Get());
        }
        else {
            fprintf(stderr, "ShapeDescriptors::ComputeShapeDescriptor unable to process %s!\n", descriptorName.c_str());
            exit(-1);
        }
    }
    return propertySetArray;
}

void ShapeDescriptors::
SplitPointCloud(PointCloudPropertySet* pSegmentation, vector<int>& segIdxArray, vector<int>& vtxIdxArray,
                vector<PointCloud*>& pointCloudArray)
{
    PointCloud* pWholeCloud = pSegmentation->GetPointCloud();
    for (int i=0; i<pSegmentation->NProperties(); i++) {
        PointCloudProperty* pSeg = pSegmentation->Property(i);
        if (pSeg->Name() == "all") continue;
        PointCloud* pSegCloud = new PointCloud;
        pSegCloud->SetName(pSeg->Name());
        for (int j=0; j<pWholeCloud->NPoints(); j++) {
            if (pSeg->VertexValue(j)<0.5) continue;
            segIdxArray[j] = (int)pointCloudArray.size();
            vtxIdxArray[j] = pSegCloud->NPoints();
            pSegCloud->AddVertex(pWholeCloud->Vertex(j)->Position(), pWholeCloud->Vertex(j)->Normal());
        }
        pointCloudArray.push_back(pSegCloud);
    }
}

vector<PointCloudPropertySet*> ShapeDescriptors::
ComputeShapeDescriptorPerSegment(R3Mesh* pMesh, PointCloud* pCloud,
                                 PointCloudPropertySet* pSegmentation,
                                 const vector<string>& descriptorNameArray)
{
    // split original mesh into different segments
    if (pSegmentation->GetPointCloud() != pCloud) {
        fprintf(stderr, "ShapeDescriptors::ComputeShapeDescriptorPerSegment pSegmentation->GetPointCloud() != pCloud!\n");
        exit(-1);
    }
    printf("ShapeDescriptors::ComputeShapeDescriptorPerSegment\n");
    printf("    Split the entire cloud ...\n");
    vector<int> segIdxArray(pCloud->NPoints(), -1);
    vector<int> vtxIdxArray(pCloud->NPoints(), -1);
    vector<PointCloud*> pointCloudArray;
    ShapeDescriptors::SplitPointCloud(pSegmentation, segIdxArray, vtxIdxArray, pointCloudArray);
    // compute shape descriptor separately
    printf("    Compute descriptors on each segment cloud individually ...\n");
    vector< vector<PointCloudPropertySet*> > prpArray;
    for (int i=0; i<(int)pointCloudArray.size(); i++) {
        printf("        cloud-name = %s\n", pointCloudArray[i]->Name().c_str());
        prpArray.push_back(ShapeDescriptors::ComputeShapeDescriptor(pMesh, pointCloudArray[i], descriptorNameArray));
    }
    // put properties of segments together
    printf("    Put descriptors together ...\n");
    vector<PointCloudPropertySet*> pWholePrpSetArray;
    // = new PointCloudPropertySet(pCloud);
    // len(prpArray) = num of parts
    // len(prpArray[0]) = num of descriptors
    int numPrps = prpArray[0].size();
    printf("        num of descriptors = %d\n", numPrps);
    for (int i=0; i<numPrps; i++) {
        PointCloudPropertySet* pCurrentPrpSet = new PointCloudPropertySet(pCloud);
        // we assume all propertySet has exactly one property
        string name = prpArray[0][i]->Property(0)->Name();
        pCurrentPrpSet->Insert(new PointCloudProperty(pCloud, name));
        for (int j=0; j<pCloud->NPoints(); j++) {
            if (segIdxArray[j] < 0 || segIdxArray[i] >= prpArray.size()) {
                fprintf(stderr, "Invalid segIdxArray[j]!\n");
                exit(-1);
            }
            if (vtxIdxArray[j] < 0 || vtxIdxArray[j] >= prpArray[segIdxArray[j]][i]->Property(0)->NVertexValues()) {
                printf("vtxIdxArray[%d] = %d! invalid!\n", j, vtxIdxArray[j]);
                exit(-1);
            }
            double value = prpArray[segIdxArray[j]][i]->Property(0)->VertexValue(vtxIdxArray[j]);
            pCurrentPrpSet->Property(0)->SetVertexValue(j, value);
        }
        pWholePrpSetArray.push_back(pCurrentPrpSet);
    }
    // clean up all auxillary point clouds and properties
    for (int i=0; i<(int)prpArray.size(); i++) {
        for (int j=0; j<(int)prpArray[i].size(); j++) {
            delete prpArray[i][j];
        }
    }
    for (int i=0; i<(int)pointCloudArray.size(); i++) delete pointCloudArray[i];
    return pWholePrpSetArray;
}
