#ifndef MY_DEFORMABLE_CYLINDER_H
#define MY_DEFORMABLE_CYLINDER_H
#include <vector>

class DeformableCircle
{
public:
    DeformableCircle(int numSamples);
    void SetRadius(double _radius);
    void SetCenter(const R3Point& pos);
    const R3Vector Normal(int idx) const;
    const R3Point Position(int idx) const;
    const R3Point& Center(void) const;
    double Radius(void) const;
    int NSamples(void) const;
protected:
    std::vector<R3Vector> m_aNormals;
    double radius;
    R3Point center;
};

class DeformableCylinder : public PointCloud {
public:
    // Constructor functions
    DeformableCylinder(int numCircles, int numSamplesPerCircle);
    virtual ~DeformableCylinder(void);
    void CoarseFit(PointCloud* pCloud);
    void FineFit(PointCloud* pCloud);
    int NCircles(void) const;
    // Skeleton
    PointCloudPropertySet* GetSkeletonProperties(const std::vector<std::string>& propertyNames) const;
    virtual void Transform(const R3Transformation& transformation);
protected:
    void UpdatePoints(void);
    // properties
    PointCloudProperty* GenerateSkeletonProperty(const std::string& prpName, PointCloud* pSkeleton) const;
    PointCloudProperty* GenerateCurvature(PointCloud* pSkeleton) const;
    PointCloudProperty* GenerateThickness(PointCloud* pSkeleton) const;
    PointCloudProperty* GenerateTapering(PointCloud* pSkeleton) const;
protected:
    std::vector<DeformableCircle*> m_aCircleArray;
};

#endif