#include "Scene.h"
using namespace std;
R2MyPixel::
R2MyPixel(void)
:x(0),y(0),
color(0, 0, 0),
owner(NULL),
normal(R2null_vector)
{}

R2MyPixel::
R2MyPixel(int _x, int _y, RNRgb _c, R3MySceneNode* _owner)
:x(_x),
y(_y),
color(_c),
owner(_owner),
normal(R2null_vector)
{}

R2MyPixel::
R2MyPixel(const R2MyPixel& _pixel)
:x(_pixel.x),
y(_pixel.y),
color(_pixel.color),
owner(_pixel.owner),
normal(_pixel.normal)
{}

int R2MyPixel::
X(void) const
{
    return x;
}

int R2MyPixel::
Y(void) const
{
    return y;
}

const R2Point R2MyPixel::
Position(void) const
{
    return R2Point(x, y);
}

const R2Vector& R2MyPixel::
Normal(void) const
{
    return normal;
}

const RNRgb& R2MyPixel::
Color(void) const
{
    return color;
}

// source: http://www.easyrgb.com/index.php?X=MATH
const R3Vector R2MyPixel::
XYZColor(void) const
{
    RNScalar var_R = color[0], var_G = color[1], var_B = color[2];
    if ( var_R > 0.04045 )
    {
        var_R = pow( (var_R + 0.055 ) / 1.055, 2.4);
    }
    else
    {
        var_R = var_R / 12.92;
    }
    if ( var_G > 0.04045 )
    {
        var_G = pow( (var_G + 0.055 ) / 1.055, 2.4);
    }
    else
    {
        var_G = var_G / 12.92;
    }
    if ( var_B > 0.04045 )
    {
        var_B = pow( (var_B + 0.055 ) / 1.055, 2.4);
    }
    else
    {
        var_B = var_B / 12.92;
    }
    var_R *= 100;
    var_G *= 100;
    var_B *= 100;
    RNScalar X, Y, Z;
    //Observer. = 2°, Illuminant = D65
    X = var_R * 0.4124 + var_G * 0.3576 + var_B * 0.1805;
    Y = var_R * 0.2126 + var_G * 0.7152 + var_B * 0.0722;
    Z = var_R * 0.0193 + var_G * 0.1192 + var_B * 0.9505;
    return R3Vector(X, Y, Z);
}

// source: http://www.easyrgb.com/index.php?X=MATH
const R3Vector R2MyPixel::
LABColor(void) const
{
    //Observer= 2°, Illuminant= D65
    R3Vector ref(95.047, 100, 108.883);
    R3Vector xyz = XYZColor();
    xyz[0] = xyz[0] / ref[0];
    xyz[1] = xyz[1] / ref[1];
    xyz[2] = xyz[2] / ref[2];
    R3Vector lab(0, 0, 0);
    lab[0] = 116*ColorForwardTransformation(xyz[1]) - 16;
    lab[1] = 500*(ColorForwardTransformation(xyz[0]) - ColorForwardTransformation(xyz[1]));
    lab[2] = 200*(ColorForwardTransformation(xyz[1]) - ColorForwardTransformation(xyz[2]));
//    R3Vector _xyz = XYZColor();
//    printf("Color [%g %g %g] -> XYZ[%g %g %g] -> Lab [%g %g %g]\n",
//           color[0], color[1], color[2],
//           _xyz[0], _xyz[1], _xyz[2],
//           lab[0], lab[1], lab[2]);
//    getchar();
    return lab;
}

RNScalar R2MyPixel::
ColorForwardTransformation(RNScalar r) const
{
    static RNScalar thres = pow(6.0/29.0, 3.0);
    static RNScalar coeff = 0.3333*(29.0*29.0/6.0/6.0);
    if (r > thres)
    {
        return pow(r, 0.3333);
    }
    else
    {
        return coeff*r + 4.0/29.0;
    }
}

R3MySceneNode* R2MyPixel::
Owner(void) const
{
    return owner;
}

RNLength R2MyPixel::
CoordinateDistance(const R2MyPixel& pixel) const
{
    RNLength dx = pixel.x - x;
    RNLength dy = pixel.y - y;
    return sqrt(dx*dx + dy*dy);
}

R2MyPixel& R2MyPixel::
operator=(const R2MyPixel& _pixel)
{
    if (this == &_pixel)
    {
        return *this;
    }
    x = _pixel.x;
    y = _pixel.y;
    color = _pixel.color;
    owner = _pixel.owner;
    normal = _pixel.normal;
    return *this;
}

RNBoolean R2MyPixel::
operator==(const R2MyPixel& _pixel) const
{
    if (x == _pixel.x &&
        y == _pixel.y &&
        color == _pixel.color &&
        owner == _pixel.owner)
    {
        return TRUE;
    }
    return FALSE;
}

RNBoolean R2MyPixel::
operator!=(const R2MyPixel& _pixel) const
{
    return !(*this == _pixel);
}

void R2MyPixel::
SetNormal(const R2Vector& n)
{
    normal = n;
}

void R2MyPixel::
SetPosition(int _x, int _y)
{
    x = _x;
    y = _y;
}

void R2MyPixel::
SetOwner(R3MySceneNode* _owner)
{
    owner = _owner;
}

void R2MyPixel::
SetColor(const RNRgb& _color)
{
    color = _color;
}