/* Source file for the R2My texture class */



/* Include files */
#include "R3MyGraphics/R3MyGraphics.h"



/* Public variables */

R2MyTexture R2Mynull_texture;



/* Public functions */

int 
R2MyInitTexture()
{
  /* Return success */
  return TRUE;
}



void 
R2MyStopTexture()
{
}



R2MyTexture::
R2MyTexture(void)
  : image(NULL),
    s_wrap(R2My_NO_TEXTURE_WRAP),
    t_wrap(R2My_NO_TEXTURE_WRAP),
    min_filter(R2My_NO_TEXTURE_FILTER),
    mag_filter(R2My_NO_TEXTURE_FILTER),
    blend(R2My_NO_TEXTURE_BLEND),    
    flags(0),
    id(0)
{
}



R2MyTexture::
R2MyTexture(const R2MyTexture& texture)
  : image(texture.image),
    s_wrap(texture.s_wrap),
    t_wrap(texture.t_wrap),
    min_filter(texture.min_filter),
    mag_filter(texture.mag_filter),
    blend(texture.blend),
    flags(texture.flags),
    id((texture.image) ? -1 : 0)
{
    // Update texture                                                                                                   
    Update();
}



R2MyTexture::
R2MyTexture(const R2MyImage *image,
  R2MyTextureWrap s_wrap, R2MyTextureWrap t_wrap,
  R2MyTextureFilter min_filter, R2MyTextureFilter mag_filter,
  R2MyTextureBlend blend)
  : image(image),
    s_wrap(s_wrap),
    t_wrap(t_wrap),
    min_filter(min_filter),
    mag_filter(mag_filter),
    blend(blend),
    flags(RN_NO_FLAGS),
    id(-1)
{
  // Update texture
  Update();
}




R2MyTexture::
R2MyTexture(const char *filename,
  R2MyTextureWrap s_wrap, R2MyTextureWrap t_wrap,
  R2MyTextureFilter min_filter, R2MyTextureFilter mag_filter,
  R2MyTextureBlend blend)
  : image(NULL),
    s_wrap(s_wrap),
    t_wrap(t_wrap),
    min_filter(min_filter),
    mag_filter(mag_filter),
    blend(blend),
    flags(RN_NO_FLAGS),
    id(-1)
{
  // Create image
  image = new R2MyImage(filename);
  assert(image);

  // Update 
  if (!image) id = 0;
  Update();
}



R2MyTexture::
~R2MyTexture(void)
{
  // Unload texture
  if (IsLoaded() > 0) Unload();
}



void R2MyTexture::
SetImage(const R2MyImage *image)
{
  // Unload previous texture
  if (IsLoaded()) Unload();

  // Set image
  flags.Remove(R2My_TEXTURE_TRANSPARENCY_FLAG);
  if ((image) && ((image->NComponents() == 2) || (image->NComponents() == 4)))
    flags.Add(R2My_TEXTURE_TRANSPARENCY_FLAG);
  this->image = image;
  this->id = -1;
}



void R2MyTexture::
Update (void)
{
  // Update flags
  UpdateFlags(RN_NO_FLAGS);
}



void R2MyTexture::
UpdateFlags (const RNFlags flags)
{
  // Update flags
  this->flags = flags;
  if ((image) && ((image->NComponents() == 2) || (image->NComponents() == 4)))
    this->flags.Add(R2My_TEXTURE_TRANSPARENCY_FLAG);
}



void R2MyTexture::
Load(void) const
{
  // Check if needs to be loaded
  if (id >= 0) return;

  // Create texture id
  GLuint i;
  glGenTextures(1, &i);
  ((R2MyTexture *) this)->id = i;
  assert(id > 0);

  // Begin texture definition                                                                                     
  glBindTexture(GL_TEXTURE_2D, id);

  // Define texture parameters
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, s_wrap);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, t_wrap);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, blend);

  // Determine texture format
  GLenum format = GL_RGB;
  if (image->Depth() == 1) format = GL_LUMINANCE;
  else if (image->Depth() == 2) format = GL_LUMINANCE_ALPHA;
  else if (image->Depth() == 3) format = GL_RGB;
  else if (image->Depth() == 4) format = GL_RGBA;
  else RNAbort("Illegal texture image");

  // Define texture
  if ((min_filter == R2My_NEAREST_MIPMAP_NEAREST_TEXTURE_FILTER) ||
      (min_filter == R2My_NEAREST_MIPMAP_LINEAR_TEXTURE_FILTER) ||
      (min_filter == R2My_LINEAR_MIPMAP_NEAREST_TEXTURE_FILTER) ||
      (min_filter == R2My_LINEAR_MIPMAP_LINEAR_TEXTURE_FILTER)) {
    gluBuild2DMipmaps(GL_TEXTURE_2D, image->Depth(), image->Width(), image->Height(),
      format, GL_UNSIGNED_BYTE, (const unsigned char *) image->Pixels());
  }
  else {
    glTexImage2D(GL_TEXTURE_2D, 0, image->Depth(), image->Width(), image->Height(),
      0, format, GL_UNSIGNED_BYTE, (const unsigned char *) image->Pixels());
  }

  // End texture definition                                                                                       
  glBindTexture(GL_TEXTURE_2D, 0);
}



void R2MyTexture::
Unload(void) const
{
  // Check id
  if (!IsLoaded()) return;

  // Delete texture
  GLuint i = id;
  glDeleteTextures(1, &i);

  // Reset texture id
  ((R2MyTexture *) this)->id = -1;
}



void R2MyTexture::
Draw(RNBoolean force) const
{
  // Check if same texture as last time
  static const R2MyTexture *R2Mycurrent_texture = NULL;
  if (!force && (this == R2Mycurrent_texture)) return;

  // Check if null texture
  if (id == 0) {
    // Disable texture
    glDisable(GL_TEXTURE_2D);
  }
  else {
    // Load texture
    if (!IsLoaded()) Load();

    // Set texture
    glBindTexture(GL_TEXTURE_2D, id);

    // Enable texture
    glEnable(GL_TEXTURE_2D);
  }

  // Remember current texture
  R2Mycurrent_texture = this;
}














  
