#ifndef R2_MY_TEXTURE_H
#define R2_MY_TEXTURE_H

/* Initialization functions */

int R2MyInitTexture();
void R2MyStopTexture();



/* Type definitions */

typedef enum {
  R2My_NO_TEXTURE_WRAP = GL_REPEAT,
  R2My_REPEAT_TEXTURE_WRAP = GL_REPEAT,
  R2My_CLAMP_TEXTURE_WRAP = GL_CLAMP
} R2MyTextureWrap;

typedef enum {
  R2My_NO_TEXTURE_FILTER = GL_NEAREST,
  R2My_NEAREST_TEXTURE_FILTER = GL_NEAREST,
  R2My_NEAREST_MIPMAP_NEAREST_TEXTURE_FILTER = GL_NEAREST_MIPMAP_NEAREST,
  R2My_NEAREST_MIPMAP_LINEAR_TEXTURE_FILTER = GL_NEAREST_MIPMAP_LINEAR,
  R2My_LINEAR_TEXTURE_FILTER = GL_LINEAR,
  R2My_LINEAR_MIPMAP_NEAREST_TEXTURE_FILTER = GL_LINEAR_MIPMAP_NEAREST,
  R2My_LINEAR_MIPMAP_LINEAR_TEXTURE_FILTER = GL_LINEAR_MIPMAP_LINEAR
} R2MyTextureFilter;

typedef enum {
  R2My_NO_TEXTURE_BLEND = GL_MODULATE,
  R2My_MODULATE_TEXTURE_BLEND = GL_MODULATE,
  R2My_DECAL_TEXTURE_BLEND = GL_DECAL,
  R2My_BLEND_TEXTURE_BLEND = GL_BLEND
} R2MyTextureBlend;



/* Class definition */

class R2MyTexture {
public:
  // Constructor functions
  R2MyTexture(void);
  R2MyTexture(const R2MyTexture& texture);
  R2MyTexture(const R2MyImage *image,
            R2MyTextureWrap s_wrap = R2My_REPEAT_TEXTURE_WRAP,
            R2MyTextureWrap t_wrap = R2My_REPEAT_TEXTURE_WRAP,
            R2MyTextureFilter min_filter = R2My_LINEAR_TEXTURE_FILTER,
            R2MyTextureFilter mag_filter = R2My_LINEAR_TEXTURE_FILTER,
            R2MyTextureBlend blend = R2My_MODULATE_TEXTURE_BLEND);
  R2MyTexture(const char *filename,
            R2MyTextureWrap s_wrap = R2My_CLAMP_TEXTURE_WRAP,
            R2MyTextureWrap t_wrap = R2My_CLAMP_TEXTURE_WRAP,
            R2MyTextureFilter min_filter = R2My_LINEAR_TEXTURE_FILTER,
            R2MyTextureFilter mag_filter = R2My_LINEAR_TEXTURE_FILTER,
            R2MyTextureBlend blend = R2My_MODULATE_TEXTURE_BLEND);
  ~R2MyTexture(void);

  // Property functions/operators
  const R2MyImage *Image(void) const;
  const R2MyTextureWrap SWrap(void) const;
  const R2MyTextureWrap TWrap(void) const;
  const R2MyTextureFilter MinFilter(void) const;
  const R2MyTextureFilter MagFilter(void) const;
  const R2MyTextureBlend Blend(void) const;
  const RNBoolean IsTransparent(void) const;
  const RNBoolean IsLoaded(void) const;
  const RNFlags Flags(void) const;
  const int ID(void) const;

  // Manipulation functions/operations
  void SetImage(const R2MyImage *image);

  // Draw functions/operations
  void Load(void) const;
  void Unload(void) const;
  void Draw(RNBoolean force = FALSE) const;

protected:
  // Upkeep functions/operators
  void Update(void);
  void UpdateFlags(const RNFlags flags);

private:
  const R2MyImage *image;
  R2MyTextureWrap s_wrap, t_wrap;
  R2MyTextureFilter min_filter, mag_filter;
  R2MyTextureBlend blend;
  RNFlags flags;
  int id; // 0=notexture, <0=unloaded, >0=loaded
};



/* Flag mask definitions */

#define R2My_TEXTURE_FLAGS                0x00000700
#define R2My_TEXTURE_TRANSPARENCY_FLAG    0x00000100



/* Public variables */

extern R2MyTexture R2Mynull_texture;
#define R2Mydefault_texture R2Mynull_texture



/* Inline functions */

inline const R2MyImage *R2MyTexture::
Image(void) const
{
  // Return image
  return image;
}



inline const R2MyTextureWrap R2MyTexture::
SWrap(void) const
{
  // Return wrap mode for s coordinate                                                                                
  return s_wrap;
}



inline const R2MyTextureWrap R2MyTexture::
TWrap(void) const
{
  // Return wrap mode for t coordinate                                                                                
  return t_wrap;
}



inline const R2MyTextureFilter R2MyTexture::
MinFilter(void) const
{
  // Return minification filter                                                                                       
  return min_filter;
}



inline const R2MyTextureFilter R2MyTexture::
MagFilter(void) const
{
  // Return magnification filter                                                                                      
  return mag_filter;
}



inline const R2MyTextureBlend R2MyTexture::
Blend(void) const
{
  // Return blending mode                                                                                             
  return blend;
}



inline const RNBoolean R2MyTexture::
IsTransparent(void) const
{
  // Return whether has transparency
  return flags[R2My_TEXTURE_TRANSPARENCY_FLAG];
}



inline const RNBoolean R2MyTexture::
IsLoaded(void) const
{
  // Return whether texture has been loaded to graphics card for rendering
  return (id > 0) ? TRUE : FALSE;
}



inline const RNFlags R2MyTexture::
Flags(void) const
{
  // Return flags
  return flags;
}



inline const int R2MyTexture::
ID(void) const
{
  // Return id
  return id;
}
#endif



