/* Source file for the R3 material class */



/* Include files */
#include "R3MyGraphics/R3MyGraphics.h"



/* Public variables */

R3MyMaterial R3Mynull_material;
R3MyMaterial R3Myblack_material;
R3MyMaterial R3Myred_material;
R3MyMaterial R3Mygreen_material;
R3MyMaterial R3Myblue_material;
R3MyMaterial R3Myyellow_material;
R3MyMaterial R3Mycyan_material;
R3MyMaterial R3Mymagenta_material;
R3MyMaterial R3Mywhite_material;
R3MyMaterial R3Mygray_material;
R3MyMaterial R3Myshiny_black_material;
R3MyMaterial R3Myshiny_red_material;
R3MyMaterial R3Myshiny_green_material;
R3MyMaterial R3Myshiny_blue_material;
R3MyMaterial R3Myshiny_yellow_material;
R3MyMaterial R3Myshiny_cyan_material;
R3MyMaterial R3Myshiny_magenta_material;
R3MyMaterial R3Myshiny_white_material;
R3MyMaterial R3Myshiny_gray_material;
R3MyMaterial R3Mydefault_material;



/* Public functions */

int 
R3InitMaterial()
{
    /* Initialize public variables */
    R3Mynull_material = R3MyMaterial(&R3null_brdf, "Null");
    R3Myblack_material = R3MyMaterial(&R3black_brdf, "Black");
    R3Myred_material = R3MyMaterial(&R3red_brdf, "Red");
    R3Mygreen_material = R3MyMaterial(&R3green_brdf, "Green");
    R3Myblue_material = R3MyMaterial(&R3blue_brdf, "Blue");
    R3Myyellow_material = R3MyMaterial(&R3yellow_brdf, "Yellow");
    R3Mycyan_material = R3MyMaterial(&R3cyan_brdf, "Cyan");
    R3Mymagenta_material = R3MyMaterial(&R3magenta_brdf, "Magenta");
    R3Mywhite_material = R3MyMaterial(&R3white_brdf, "White");
    R3Mygray_material = R3MyMaterial(&R3gray_brdf, "Gray");
    R3Myshiny_black_material = R3MyMaterial(&R3shiny_black_brdf, "ShinyBlack");
    R3Myshiny_red_material = R3MyMaterial(&R3shiny_red_brdf, "ShinyRed");
    R3Myshiny_green_material = R3MyMaterial(&R3shiny_green_brdf, "ShinyGreen");
    R3Myshiny_blue_material = R3MyMaterial(&R3shiny_blue_brdf, "ShinyBlue");
    R3Myshiny_yellow_material = R3MyMaterial(&R3shiny_yellow_brdf, "ShinyYellow");
    R3Myshiny_cyan_material = R3MyMaterial(&R3shiny_cyan_brdf, "ShinyCyan");
    R3Myshiny_magenta_material = R3MyMaterial(&R3shiny_magenta_brdf, "ShinyMagenta");
    R3Myshiny_white_material = R3MyMaterial(&R3shiny_white_brdf, "ShinyWhite");
    R3Myshiny_gray_material = R3MyMaterial(&R3shiny_gray_brdf, "ShinyGray");
    R3Mydefault_material = R3MyMaterial(&R3default_brdf, "Default");
    
    /* Return success */
    return TRUE;
}



void 
R3StopMaterial()
{
}



R3MyMaterial::
R3MyMaterial(const char *name)
    : brdf(NULL),
      texture(NULL),
      flags(0),
      name(NULL)
{
    // Set name
    if (name) this->name = strdup(name);

    // Update flags
    Update();
}



R3MyMaterial::
R3MyMaterial(const R3MyMaterial& material)
    : brdf(material.brdf),
      texture(material.texture),
      flags(material.flags),
      name(NULL)
{
    // Set name
    if (material.name) this->name = strdup(material.name);
}



R3MyMaterial::
R3MyMaterial(const R3Brdf *brdf, const char *name)
    : brdf(brdf),
      texture(&R2Mynull_texture),
      flags(0),
      name(NULL)
{
    // Set name
    if (name) this->name = strdup(name);

    // Update flags
    Update();
}



R3MyMaterial::
R3MyMaterial(const R2MyTexture *texture, const char *name)
    : brdf(&R3white_brdf),
      texture(texture),
      flags(0),
      name(NULL)
{
    // Set name
    if (name) this->name = strdup(name);

    // Update flags
    Update();
}



R3MyMaterial::
R3MyMaterial(const R3Brdf *brdf, const R2MyTexture *texture, const char *name)
    : brdf(brdf),
      texture(texture),
      flags(0),
      name(NULL)
{
    // Set name
    if (name) this->name = strdup(name);

    // Update flags
    Update();
}



R3MyMaterial::
~R3MyMaterial(void)
{
    // Free name
    if (name) free(name);
}


 
R3MyMaterial& R3MyMaterial::
operator=(const R3MyMaterial& material)
{
    // Update everything
    SetBrdf(material.brdf);
    SetTexture(material.texture);
    SetName(material.name);
    Update();
    return *this;
}



void R3MyMaterial::
SetBrdf(const R3Brdf *brdf)
{
    // Set brdf
    if ((this->brdf) && (!brdf)) flags.Remove(R3_MATERIAL_BRDF_FLAG);
    if ((!this->brdf) && (brdf)) flags.Add(R3_MATERIAL_BRDF_FLAG);
    if (this->brdf) flags.Remove(this->brdf->Flags());
    if (brdf) flags.Add(brdf->Flags());
    this->brdf = brdf;
}




void R3MyMaterial::
SetTexture(const R2MyTexture *texture)
{
    // Set texture
    if ((this->texture) && (!texture)) flags.Remove(R3_MATERIAL_TEXTURE_FLAG);
    if ((!this->texture) && (texture)) flags.Add(R3_MATERIAL_TEXTURE_FLAG);
    if (this->texture) flags.Remove(this->texture->Flags());
    if (texture) flags.Add(texture->Flags());
    this->texture = texture;
}



void R3MyMaterial::
SetName(const char *name)
{
    // Set name
    if (this->name) free(this->name);
    if (name) this->name = strdup(name);
    else this->name = NULL;
}



void R3MyMaterial::
Update(void)
{
    // Update flags
    this->flags = RN_NO_FLAGS;
    if ((brdf) && (brdf->ID() != 0)) 
	this->flags.Add(R3_MATERIAL_BRDF_FLAG | brdf->Flags());
    if ((texture) && (texture->ID() != 0)) 
	this->flags.Add(R3_MATERIAL_TEXTURE_FLAG | texture->Flags());
}




void R3MyMaterial::
Load(void) const
{
    // Draw brdf
    if (brdf) brdf->Load();

    // Draw texture
    if (texture) texture->Load();
}



void R3MyMaterial::
Unload(void) const
{
    // Draw brdf
    if (brdf) brdf->Unload();

    // Draw texture
    if (texture) texture->Unload();
}



void R3MyMaterial::
Draw(RNBoolean force) const
{
	force = 1;
    // Check if same material as last time
    static const R3MyMaterial *R3current_material = NULL;
    if (!force && (this == R3current_material)) {
		printf("The same material?\n");
		getchar();
		return;
	}
//    // Draw brdf
//    if (brdf) brdf->Draw(force);
//    else R3default_brdf.Draw(force);
//
//    // Draw texture
//    if (texture) texture->Draw(force);
//    else R2null_texture.Draw(force);

    if (texture) {
        R3default_brdf.Draw(force);
        texture->Draw(force);
    }
    else if (brdf) {
        brdf->Draw(force);
        R2null_texture.Draw(force);
    }
    else {
        R3default_brdf.Draw(force);
        R2null_texture.Draw(force);
    }
    
    // Remember current material
    R3current_material = this;
}























