#ifndef R3_MY_MATERIAL_H
#define R3_MY_MATERIAL_H

/* Initialization functions */

int R3InitMaterial();
void R3StopMaterial();


/* Class definition */

class R3MyMaterial /* : public R3Attribute */ {
public:
    // Constructor functions
	R3MyMaterial(const char *name = NULL);
    R3MyMaterial(const R3MyMaterial& material);
    R3MyMaterial(const R3Brdf *brdf, const char *name = NULL);
    R3MyMaterial(const R2MyTexture *texture, const char *name = NULL);
    R3MyMaterial(const R3Brdf *brdf, const R2MyTexture *texture, const char *name = NULL);
	~R3MyMaterial(void);
    
    // Material property functions/operations
    const char *Name(void) const;
    const R3Brdf *Brdf(void) const;
    const R2MyTexture *Texture(void) const;
	const RNFlags& Flags(void) const;
	const RNBoolean IsReflective(void) const;
	const RNBoolean IsAmbient(void) const;
	const RNBoolean IsDiffuse(void) const;
	const RNBoolean IsSpecular(void) const;
	const RNBoolean IsEmissive(void) const;
	const RNBoolean IsShiny(void) const;
	const RNBoolean IsTextured(void) const;
	const RNBoolean IsTransparent(void) const;
    
	// Manipulation functions/operations
	R3MyMaterial& operator=(const R3MyMaterial& material);
    void SetBrdf(const R3Brdf *brdf);
    void SetTexture(const R2MyTexture *texture);
    void SetName(const char *name);
    
	// Draw functions/operations
    void Load(void) const;
    void Unload(void) const;
    void Draw(RNBoolean force = FALSE) const;
    
    // Update functions
    void Update(void);
    
private:
    const R3Brdf *brdf;
    const R2MyTexture *texture;
	RNFlags flags;
    char *name;
};



/* Public variables */

extern R3MyMaterial R3Mynull_material;
extern R3MyMaterial R3Mydefault_material;
extern R3MyMaterial R3Myblack_material;
extern R3MyMaterial R3Myred_material;
extern R3MyMaterial R3Mygreen_material;
extern R3MyMaterial R3Myblue_material;
extern R3MyMaterial R3Myyellow_material;
extern R3MyMaterial R3Mycyan_material;
extern R3MyMaterial R3Mymagenta_material;
extern R3MyMaterial R3Mywhite_material;
extern R3MyMaterial R3Mygray_material;
extern R3MyMaterial R3Myshiny_black_material;
extern R3MyMaterial R3Myshiny_red_material;
extern R3MyMaterial R3Myshiny_green_material;
extern R3MyMaterial R3Myshiny_blue_material;
extern R3MyMaterial R3Myshiny_yellow_material;
extern R3MyMaterial R3Myshiny_cyan_material;
extern R3MyMaterial R3Myshiny_magenta_material;
extern R3MyMaterial R3Myshiny_white_material;
extern R3MyMaterial R3Myshiny_gray_material;



/* Flag mask definitions */

#define R3_MATERIAL_FLAGS		0x00000FFF
#define R3_MATERIAL_BRDF_FLAG           0x00000080
#define R3_MATERIAL_TEXTURE_FLAG	0x00000800
#define R3_MATERIAL_TRANSPARENCY_FLAG	(R3_BRDF_TRANSPARENCY_FLAG | R2_TEXTURE_TRANSPARENCY_FLAG)



/* Inline functions */

inline const R3Brdf *R3MyMaterial::
Brdf(void) const
{
    // Return brdf
    return brdf;
}



inline const R2MyTexture *R3MyMaterial::
Texture(void) const
{
    // Return texture
    return texture;
}



inline const char *R3MyMaterial::
Name(void) const
{
    // Return name
    return name;
}



inline const RNFlags& R3MyMaterial::
Flags(void) const
{
    // Return flags
    return flags;
}



inline const RNBoolean R3MyMaterial::
IsReflective(void) const
{
    // Return whether has any reflectivity
    return flags[R3_MATERIAL_BRDF_FLAG];
}



inline const RNBoolean R3MyMaterial::
IsAmbient(void) const
{
    // Return whether has ambient term
    return flags[R3_BRDF_AMBIENT_FLAG];
}



inline const RNBoolean R3MyMaterial::
IsDiffuse(void) const
{
    // Return whether has diffuse term
    return flags[R3_BRDF_DIFFUSE_FLAG];
}



inline const RNBoolean R3MyMaterial::
IsSpecular(void) const
{
    // Return whether has specular term
    return flags[R3_BRDF_SPECULAR_FLAG];
}



inline const RNBoolean R3MyMaterial::
IsEmissive(void) const
{
    // Return whether has emissive term
    return flags[R3_BRDF_EMISSION_FLAG];
}



inline const RNBoolean R3MyMaterial::
IsShiny(void) const
{
    // Return whether has shininess
    return flags[R3_BRDF_SHININESS_FLAG];
}



inline const RNBoolean R3MyMaterial::
IsTextured(void) const
{
    // Return whether has texture
    return flags[R3_MATERIAL_TEXTURE_FLAG];
}



inline const RNBoolean R3MyMaterial::
IsTransparent(void) const
{
    // Return whether has transparency
    return flags[R3_MATERIAL_TRANSPARENCY_FLAG];
}

#endif












