#ifndef MY_VOXELIZATION_H
#define MY_VOXELIZATION_H
struct R3IntBox
{
    R3IntBox(int minx, int miny, int minz, int maxx, int maxy, int maxz);
    int Volume(void) const;
    int minX, minY, minZ;
    int maxX, maxY, maxZ;
};

class Voxelization
{
public:
    Voxelization(void);
    // I/O functions
    int ReadFile(const char* filename);
    int WriteFile(const char* filename) const;
    // import from mesh
    int Import(R3Mesh* pMesh, double voxelSize);
    int Import(PointCloud* pCloud, double voxelSize);
    // flags
    bool IsEmpty(void) const;
    // draw
    void Draw(void) const;
    // cuboids related
    void InitializeCuboids(double occupyRate, double stopOccupyRatio);
    // void SelectCuboids(double stopVolumeFraction, double minOverlapFraction);
    const std::vector<R3Box> GetCuboids(void) const;
protected:
    void Set(const R3Box& box, double voxelSize);
    void SetOccupyFlag(int x, int y, int z);
    void MarkOccupyFlagForCuboid(const R3IntBox& cuboid, int flag);
    const std::vector<int> GetNumberOfUnoccupiedCellsPerCuboids(void) const;
    const R3Box GetCuboid(const R3IntBox& cuboid) const;
    void MarkOccupyFlag(int x, int y, int z, int flag);
    void FloodFill(int x, int y, int z);
    int GetOccupyFlag(int x, int y, int z) const;
    void SetInteriorOccupancy(void);
    void InitializeOccupyFlags(void);
    const R3Box GetCellBox(int x, int y, int z) const;
    // find the first marked flag
    int FindMarkedItem(int flag, int startIndex=0) const;
    const R3IntBox GrowCuboid(int index, double occupyRate);
    void ConvertIndexTo3DIndices(int index, int* px, int* py, int* pz) const;
    void GetChangeOfMove(int minX, int minY, int minZ, int maxX, int maxY, int maxZ,
                         int expansionDir, int* pNewOccupy, int* pNewTotal) const;
protected:
    int m_iXRange, m_iYRange, m_iZRange;
    double m_fVoxelSize;
    double m_fXMin, m_fYMin, m_fZMin;
    std::vector<int> m_aOccupyFlags;
    std::vector<R3IntBox> m_aCuboidArray;
};
#endif