import sys
import os
import glob
import ConfigParser
import utils
import csv
import optimization
import numpy
from collections import OrderedDict

# ranking function
# first column is the ref image, and next columns are ordered by similarity, from most similar to least similar
def rank_by_similarity(descriptorArray, w, rankingFilename):
    refMesh2ValueArray = {}
    for refMshName in descriptorArray:
        refDesc = descriptorArray[refMshName]
        valueDict = {} 
        for optMshName in descriptorArray:
            optDesc = descriptorArray[optMshName]
            dist = refDesc - optDesc
            valueDict[optMshName] = optimization.similarity(dist, w)
        # sort valueDict
        valueDictSortedByValue = OrderedDict(sorted(valueDict.items(), key=lambda x: x[1]))
        # write a row to csv
        valueArray = []
        valueArray.append(refMshName)
        for mshName in valueDictSortedByValue:
            # optMshName followed by its value
            valueArray.append(mshName)
            valueArray.append(valueDictSortedByValue[mshName])
        refMesh2ValueArray[refMshName] = valueArray
    # sort by key
    sortedRefMesh2ValueArray = OrderedDict(sorted(refMesh2ValueArray.items(), key=lambda x: x[0]))
    # write to csv (can be ordered by the key?)
    fout = open(rankingFilename, 'w')
    csvwrite = csv.writer(fout)
    for refMshName in sortedRefMesh2ValueArray:
        csvwrite.writerow(sortedRefMesh2ValueArray[refMshName])
    fout.close()

# evaluation function
def evaluate_ranking(rankingFilename, evaluationFilename, gtFilename):
    # parse ranking filename
    fin = open(rankingFilename, 'r')
    csvreader = csv.reader(fin)
    ranking = {}
    for row in csvreader:
        # form a mapping from mshName to order
        mapping = {}
        count = 0
        for item in row:
            mapping[item] = count
            count = count + 1
        ranking[row[0]] = mapping
    fin.close()
    # given ranking let's see what fraction of gt is correctly predicted
    fin = open(gtFilename, 'r')
    evaluation = []
    csvreader = csv.reader(fin)
    numCorrect = 0
    numAll = 0
    for row in csvreader:
        numAll = numAll + 1
        refMshName = row[0]
        opt1MshName = row[1]
        opt2MshName = row[2]
        gtRelation = row[3]
        # what is the predicted relation according to ranking
        predRelation = '<'
        if ranking[refMshName][opt1MshName] < ranking[refMshName][opt2MshName]:
            predRelation = '<'
        else:
            predRelation = '>'
        # whether the prediction is correct?
        if predRelation == gtRelation:
            row.append('correct')
            numCorrect = numCorrect + 1
        else:
            row.append('wrong')
        # update row[3] with the predicted relation
        row[3] = predRelation
        evaluation.append(row)
    fin.close()
    # write to evaluation
    fout = open(evaluationFilename, 'w')
    csvwriter = csv.writer(fout)
    for row in evaluation:
        csvwriter.writerow(row)
    # also output correct, all, correct fraction
    summary = []
    summary.append('accuracy')
    summary.append(str(numCorrect))
    summary.append(str(numAll))
    summary.append(str(float(numCorrect)/numAll))
    csvwriter.writerow(summary)
    fout.close()
    print evaluationFilename+' %d/%d are correct!'%(numCorrect, numAll)
    
def evaluate_comparison(result, summaryFilename, gtFilename):
    fin = open(gtFilename, 'r')
    csvreader = csv.reader(fin)
    gtRelationArray = {}
    for row in csvreader:
        gtRelationArray[row[0]+','+row[1]+','+row[2]] = row[3]
    fin.close()
    # given gtFilename let's see what fraction of result is correctly predicted
    fin = open(result, 'r')
    evaluation = []
    csvreader = csv.reader(fin)
    numCorrect = 0
    numAll = 0
    for row in csvreader:
        if row[0] == 'accuracy':
            continue
        numAll = numAll + 1
        refMshName = row[0]
        opt1MshName = row[1]
        opt2MshName = row[2]
        predRelation = row[3]
        # find gtRelationArray[refMshName],opt1MshName,opt2MshName]
        query = refMshName+','+opt1MshName+','+opt2MshName
        if not query in gtRelationArray:
            print 'evaluate_comparison unable to find '+query+' in '+ gtFilename
            sys.exit()
        gtRelation = gtRelationArray[query]
        # whether the prediction is correct?
        if predRelation == gtRelation:
            row[4] = 'correct'
            numCorrect = numCorrect + 1
        else:
            row[4] = 'wrong'
        evaluation.append(row)
    fin.close()
    # write to evaluation
    fout = open(summaryFilename, 'w')
    csvwriter = csv.writer(fout)
    for row in evaluation:
        csvwriter.writerow(row)
    # also output correct, all, correct fraction
    summary = []
    summary.append('accuracy')
    summary.append(str(numCorrect))
    summary.append(str(numAll))
    summary.append(str(float(numCorrect)/numAll))
    csvwriter.writerow(summary)
    fout.close()
    print summaryFilename+' %d/%d are correct!'%(numCorrect, numAll)

def evaluate_testing(testingInDesc, testingPairnames, w, isRandom, summaryFilename):
    plainTestTripletArray = []
    correctCount = 0
    numTriplets = len(testingInDesc)
    for index in range(numTriplets):
        testTriplet = testingInDesc[index]
        if isRandom:
            wThis = numpy.random.random((w.shape[0], w.shape[1]))
            wThis = numpy.matrix(wThis)
        else:
            wThis = w
        distCmp = optimization.similarity_comparison(testTriplet[0], testTriplet[1], wThis)
        plainTestTriplet = list(testingPairnames[index])
        plainTestTriplet[3] = optimization.get_relation_from_value(distCmp)
        if plainTestTriplet[3] == testingPairnames[index][3]:
            comparisonCorrectness = 'correct'
            correctCount = correctCount + 1
        else:
            comparisonCorrectness = 'wrong'
        plainTestTriplet.append(comparisonCorrectness)
        plainTestTriplet.append(str(optimization.similarity(testTriplet[0], wThis)))
        plainTestTriplet.append(str(optimization.similarity(testTriplet[1], wThis)))
        plainTestTripletArray.append(plainTestTriplet)
    fout = open(summaryFilename, 'w')
    for line in plainTestTripletArray:
        fout.write(','.join(line)+'\n')
    fout.write('accuracy,%d,%d,%g\n'%(correctCount, numTriplets, float(correctCount)/numTriplets))
    fout.close()
    return correctCount


