import os
import sys
import glob
import ConfigParser
import platform
from pca_utils import *

#if len(sys.argv)<10:
#    print 'Usage: '+sys.argv[0]+' mesh-filename(*.off) points-filename(*.pts) segmentation-filename(*.arff) training-points-dirname training-segmentation-dirname training-surface-descriptor-reponame training-histograms-reponame workspace output-feature-filename'
#    sys.exit()

if len(sys.argv)<9:
    print 'Usage: '+sys.argv[0]+' mesh-filename(*.off) points-filename(*.pts) segmentation-filename(*.arff) data-reponame(e.g. public_data) scene-class object-class workspace output-feature-filename'
    sys.exit()

meshFilename = sys.argv[1]
ptsFilename = sys.argv[2]
segFilename = sys.argv[3]
dataReponame = sys.argv[4]
sceneClassname = sys.argv[5]
objectClassname = sys.argv[6]
trainingPtsDirname = dataReponame+'/'+sceneClassname+'/intermediate_results/samples/'+objectClassname
trainingSegDirname = dataReponame+'/'+sceneClassname+'/intermediate_results/segmentation/'+objectClassname
trainingDescReponame = dataReponame+'/'+sceneClassname+'/intermediate_results/surface-descriptors/'+objectClassname
trainingHistReponame = dataReponame+'/'+sceneClassname+'/intermediate_results/histograms/'+objectClassname
workspace = sys.argv[7]
outputFeatureFilename = sys.argv[8]

config = ConfigParser.ConfigParser()
config.read('configs/config.cfg')


print '----------------- step 1: generating surface descriptors ---------------'
descDirname = workspace+'/surface-descriptors'
os.system('mkdir -p '+descDirname)
cmd = ['code/bin/'+platform.machine()+'/gen_shape_descriptors', 
       '-Task', 'task', 'GenerateOneArff',
       '-GenerateOneArffInput', 'mesh_filename', meshFilename,
       '-GenerateOneArffInput', 'point_cloud_filename', ptsFilename, 
       '-GenerateOneArffInput', 'segmentation_filename', segFilename,
       '-GenerateOneArffInput', 'descriptor_name_array', config.get('ShapeDescriptors', 'type'),
       '-GenerateOneArffInput', 'process_descriptor_name_array_flag', 'on',
       '-GenerateOneArffOutput', 'property_dirname', descDirname]
os.system(' '.join(cmd))

print '----------------- step 2: generating histograms -------------------------'
histogramReponame = workspace+'/histograms'
descriptorTypes = config.get('ShapeDescriptors', 'type').split(',')
# temporary file for range
rangeDirname = workspace+'/ranges'
os.system('mkdir -p '+rangeDirname)
for descType in descriptorTypes:
    print descType
    # generating ranges
    rangeDescriptionFilename = rangeDirname+'/'+descType+'.txt'
    cmd = ['code/bin/'+platform.machine()+'/gen_shape_descriptors',
           '-Task', 'task', 'GenerateRangePerArffs',
           '-GenerateRangePerArffsInput', 'point_cloud_dirname', trainingPtsDirname,
           '-GenerateRangePerArffsInput', 'property_dirname', trainingDescReponame+'/'+descType,
           '-GenerateRangePerArffsInput', 'segmentation_dirname', trainingSegDirname,
           '-GenerateRangePerArffsOutput', 'range_description_filename', rangeDescriptionFilename]
    os.system(' '.join(cmd))
    # generating histograms
    os.system('mkdir -p '+histogramReponame+'/'+descType)
    cmd = ['code/bin/'+platform.machine()+'/gen_shape_descriptors', 
           '-GenerateHistogramInput', 'point_cloud_filename', ptsFilename,
           '-GenerateHistogramInput', 'segmentation_filename', segFilename,
           '-GenerateHistogramInput', 'range_description_filename', rangeDescriptionFilename,
           '-GenerateHistogramInput', 'property_filename', descDirname+'/'+descType+'.arff',
           '-GenerateHistogramOutput', 'histogram_dirname', histogramReponame+'/'+descType,
           '-Task', 'task', 'GenerateHistogram']
    os.system(' '.join(cmd))

print '----------------- step 3: generating area fraction and bbox dimensions ---------------'
cmd = ['code/bin/'+platform.machine()+'/gen_seg_descriptors',
       '-Input', 'point_cloud_filename', ptsFilename,
       '-Input', 'segmentation_filename', segFilename,
       '-Output', 'shape_descriptor_dirname', histogramReponame+'/area_fraction',
       '-DescriptorType', 'type', 'area_fraction']
os.system(' '.join(cmd))
cmd = ['code/bin/'+platform.machine()+'/gen_seg_descriptors',
       '-Input', 'point_cloud_filename', ptsFilename,
       '-Input', 'segmentation_filename', segFilename,
       '-Output', 'shape_descriptor_dirname', histogramReponame+'/bbox_dimensions',
       '-DescriptorType', 'type', 'bbox_dimensions']
os.system(' '.join(cmd))

print '----------------- step 4: run PCA ---------------'

# config parser
descTypeArray = config.get('CompressShapeDescriptors', 'type').split(',')
dims = int(config.get('CompressShapeDescriptors', 'pca_remaining_number_of_dims'))

# meshNameArray is array<string>
# partNameArray is array<string>
# descValueArray is array< array< array<nparray> > >: indexed by meshName, descType and partName respectively
# partPresenceArray is array<nparray>: indexed by meshName, nparray represents if a part exists for the shape
meshNameArray,partNameArray,descValueArray,partPresenceArray = read_all_parts_descriptor_values(trainingHistReponame, descTypeArray)
# compute mean for each histogram
# meanValues is array< array<nparray> >, indexed by descType and partName
meanValues = compute_means(descValueArray, partPresenceArray)
# fill back mean for missing fields
descValueArray = fill_missing_fields(descValueArray, partPresenceArray, meanValues, descTypeArray)
# concatenate values of various parts from the same descriptor
# now descValueArray is array< array<nparray> >
descValueArray = concatenate_values_for_parts(descValueArray)
# split descValueArray to pca part and unpca part
# both are array< array<nparray> >
pcaValueArray,noPcaValueArray = split_based_on_pca(descValueArray, descTypeArray, descTypeArray)

# merge if necessary
pcaValueArray = concatenate_values_for_desctypes(pcaValueArray)

# read originalDesc
testDescValueArray, testPartPresenceArray = read_all_parts_descriptor_values_for_single_shape(histogramReponame, descTypeArray, partNameArray)
testDescValueArray = fill_missing_fields(testDescValueArray, testPartPresenceArray, meanValues, descTypeArray)
testDescValueArray = concatenate_values_for_parts(testDescValueArray)
testPcaValueArray,testNoPcaValueArray = split_based_on_pca(testDescValueArray, descTypeArray, descTypeArray)
testPcaValueArray = concatenate_values_for_desctypes(testPcaValueArray)
# conduct pca and perform transformation
testPcaValueArray = pca_transform(testPcaValueArray, pcaValueArray, dims, False)
# final merge
valueArray = merge_values(testPcaValueArray, testNoPcaValueArray)
# write newDesc
write_descriptor_values_for_single_shape(valueArray[0], outputFeatureFilename)
