import sys
import os
import time
import optimization
import evaluation
import ConfigParser
import utils
import triplets_parser
import glob
import shape_descriptor
import numpy
import random as myrandom
from sklearn import ensemble
from sklearn import datasets

def get_modelname(filename):
    basename = os.path.basename(filename)
    return os.path.splitext(basename)[0]

if len(sys.argv)<9:
    print 'Usage: '+sys.argv[0]+' training-triplets-dirname training-descriptor-dirname reference-feature-filename opt1-feature-filename opt2-feature-filename ref-categoryname opt-categoryname workspace'
    sys.exit()

trainingTripletsDirname = sys.argv[1]
trainingFeatureDirname = sys.argv[2]
refSdFilename = sys.argv[3]
opt1SdFilename = sys.argv[4]
opt2SdFilename = sys.argv[5]
refCategoryname = sys.argv[6]
optCategoryname = sys.argv[7]
workspace = sys.argv[8]


refModelname = get_modelname(refSdFilename)
opt1Modelname = get_modelname(opt1SdFilename)
opt2Modelname = get_modelname(opt2SdFilename)

# configuration
config = ConfigParser.ConfigParser()
config.read('configs/config.cfg')

# the current index
def pick_testing_training_examples_maximally(trainingCmpResults, testingCmpResults, iterationIdx):
    testingTripletArray = []
    trainingTripletArray = []
    if iterationIdx >= len(testingCmpResults):
        return trainingTripletArray,testingTripletArray
    testingTriplet = testingCmpResults[iterationIdx]
    testingTripletArray.append(testingTriplet)
    occupiedNames = testingTriplet[:3]
    # pick all triplets in trainingCmpResults that doesn't overlap with testingTriplet
    for trainingCmpResult in trainingCmpResults:
        if trainingCmpResult[0] in occupiedNames or trainingCmpResult[1] in occupiedNames or trainingCmpResult[2] in occupiedNames:
            continue
        trainingTripletArray.append(trainingCmpResult)
    return trainingTripletArray,testingTripletArray

def retrieve_model_feature(sdFilename):
    return shape_descriptor.read(sdFilename)

# descriptorDict is in the form of array
def retrieve_training_features(modelNameSet, featureDirname):
    # retrieve model names and descriptors
    descriptorDict = {}
    for modelName in modelNameSet:
        tempDesc = retrieve_model_feature(featureDirname+'/'+modelName+'.sd')
        descriptorDict[modelName] = []
        for idx in range(tempDesc.size):
            descriptorDict[modelName].append(float(tempDesc[idx]))
    return descriptorDict

def retrieve_testing_features(sdFilename1, sdFilename2, sdFilename3):
    descriptorDict = {}
    filenameArray = [sdFilename1, sdFilename2, sdFilename3]
    for filename in filenameArray:
        tempDesc = retrieve_model_feature(filename)
        modelname = get_modelname(filename)
        descriptorDict[modelname] = []
        for idx in range(tempDesc.size):
            descriptorDict[modelname].append(float(tempDesc[idx]))
    return descriptorDict

def get_values_from_dictionary(dictionary, keywordArray):
    valueDict = {}
    for keyword in keywordArray:
        if not keyword in dictionary:
            print 'get_values_from_dictionary unable to find keyword %s!'%keyword
            sys.exit()
        valueDict[keyword] = dictionary[keyword]
    return valueDict

def array_to_column_vector(src):
    tar = numpy.matrix(src)
    return tar.reshape(tar.size, 1)

def stack_descriptors(refDesc, optDesc, refCategoryIndex, optCategoryIndex, categoryDescDims):
    stack = numpy.matrix(numpy.empty((0,1)))
    for index in range(len(categoryDescDims)):
        if index == refCategoryIndex:
            if refDesc.size != categoryDescDims[index]:
                print 'stack_descriptor num of dims doesnt match!'
                sys.exit()
            stack = numpy.row_stack((stack, refDesc))
        elif index == optCategoryIndex:
            if optDesc.size != categoryDescDims[index]:
                print 'stack_descriptor num of dims doesnt match!'
                sys.exit()
            stack = numpy.row_stack((stack, -optDesc))
        else:
            emptyCol = numpy.matrix(numpy.zeros((categoryDescDims[index],1)))
            stack = numpy.row_stack((stack, emptyCol))
    return stack

def construct_cmp_results_in_desc_single(tripletNameArray, attrDict, categoryDescDims):
    # tripletDiffArray
    tripletDiffArray = []
    for i in range(len(tripletNameArray)):
        tripletName = tripletNameArray[i]
        refAttr = array_to_column_vector(attrDict[tripletName[0]])
        opt1Attr = array_to_column_vector(attrDict[tripletName[1]])
        opt2Attr = array_to_column_vector(attrDict[tripletName[2]])
        relation = 0
        if tripletName[3] == '>':
            relation = 0
        elif tripletName[3] == '<':
            relation = 1
        else:
            print '[Warning] unable to convert %s to a relation (only <,> are accepted!)'%tripletName[3]
            sys.exit()
        refCategoryIndex = tripletName[4]
        optCategoryIndex = tripletName[5]
        stack1 = refAttr - opt1Attr
        stack2 = refAttr - opt2Attr
        # accelerate computation of differentials
        quickSave = stack1*stack1.transpose() - stack2*stack2.transpose()
        tripletDiffArray.append([stack1, stack2, relation, quickSave])
    return tripletDiffArray

def construct_cmp_results_compress_in_desc_single(tripletNameArray, attrDict, categoryDescDims):
    # tripletDiffArray
    tripletDiffArray = []
    for i in range(len(tripletNameArray)):
        tripletName = tripletNameArray[i]
        refAttr = array_to_column_vector(attrDict[tripletName[0]])
        opt1Attr = array_to_column_vector(attrDict[tripletName[1]])
        opt2Attr = array_to_column_vector(attrDict[tripletName[2]])
        # relation array
        relationArray = tripletName[3]
        refCategoryIndex = tripletName[4]
        optCategoryIndex = tripletName[5]
        stack1 = refAttr - opt1Attr
        stack2 = refAttr - opt2Attr
        # accelerate computation of differentials
        quickSave = stack1*stack1.transpose() - stack2*stack2.transpose()
        tripletDiffArray.append([stack1, stack2, relationArray, quickSave])
    return tripletDiffArray

def construct_cmp_results_in_desc(tripletNameArray, attrDict, categoryDescDims):
    # tripletDiffArray
    tripletDiffArray = []
    for i in range(len(tripletNameArray)):
        tripletName = tripletNameArray[i]
        refAttr = array_to_column_vector(attrDict[tripletName[0]])
        opt1Attr = array_to_column_vector(attrDict[tripletName[1]])
        opt2Attr = array_to_column_vector(attrDict[tripletName[2]])
        relation = 0
        if tripletName[3] == '>':
            relation = 0
        elif tripletName[3] == '<':
            relation = 1
        else:
            print '[Warning] unable to convert %s to a relation (only <,> are accepted!)'%tripletName[3]
            sys.exit()
        refCategoryIndex = tripletName[4]
        optCategoryIndex = tripletName[5]
        stack1 = stack_descriptors(refAttr, opt1Attr, refCategoryIndex, optCategoryIndex, categoryDescDims)
        stack2 = stack_descriptors(refAttr, opt2Attr, refCategoryIndex, optCategoryIndex, categoryDescDims)
        # accelerate computation of differentials
        quickSave = stack1*stack1.transpose() - stack2*stack2.transpose()
        tripletDiffArray.append([stack1, stack2, relation, quickSave])
    return tripletDiffArray

def construct_cmp_results_compress_in_desc(tripletNameArray, attrDict, categoryDescDims):
    # tripletDiffArray
    tripletDiffArray = []
    for i in range(len(tripletNameArray)):
        tripletName = tripletNameArray[i]
        refAttr = array_to_column_vector(attrDict[tripletName[0]])
        opt1Attr = array_to_column_vector(attrDict[tripletName[1]])
        opt2Attr = array_to_column_vector(attrDict[tripletName[2]])
        # relation array
        relationArray = tripletName[3]
        refCategoryIndex = tripletName[4]
        optCategoryIndex = tripletName[5]
        stack1 = stack_descriptors(refAttr, opt1Attr, refCategoryIndex, optCategoryIndex, categoryDescDims)
        stack2 = stack_descriptors(refAttr, opt2Attr, refCategoryIndex, optCategoryIndex, categoryDescDims)
        # accelerate computation of differentials
        quickSave = stack1*stack1.transpose() - stack2*stack2.transpose()
        tripletDiffArray.append([stack1, stack2, relationArray, quickSave])
    return tripletDiffArray

def extract_desc_dims(tripletArray, attrDict):
    dimsMapping = {}
    maxCategoryIndex = -1
    for triplet in tripletArray:
        # ref
        refIndex = triplet[4]
        refName = triplet[0]
        if not refName in attrDict:
            print 'extract_desc_dims unable to find refName = '+refName+' !'
            sys.exit()
        refDesc = attrDict[refName]
        if not refIndex in dimsMapping:
            dimsMapping[refIndex] = len(refDesc)
            if refIndex>maxCategoryIndex:
                maxCategoryIndex = refIndex
        elif dimsMapping[refIndex] != len(refDesc):
            print 'extract_desc_dims dims doesnt match for category %d!'%refIndex
            sys.exit()
        # opt
        optIndex = triplet[5]
        optName1 = triplet[1]
        optName2 = triplet[2]
        if (not optName1 in attrDict) or (not optName2 in attrDict):
            print 'extract_desc_dims unable to find optName!'
            sys.exit()
        optDesc1 = attrDict[optName1]
        optDesc2 = attrDict[optName2]
        if len(optDesc1) != len(optDesc2):
            print 'extract_desc_dims len(optDesc1) != len(optDesc2)!'
            sys.exit()
        if not optIndex in dimsMapping:
            dimsMapping[optIndex] = len(optDesc1)
            if optIndex>maxCategoryIndex:
                maxCategoryIndex = optIndex
        elif dimsMapping[optIndex] != len(optDesc1):
            print 'extract_desc_dims dims doesnt match for category %d!'%optIndex
            sys.exit()
    descDims = []
    for index in range(maxCategoryIndex+1):
        if index in dimsMapping:
            descDims.append(dimsMapping[index])
        else:
            descDims.append(0)
    return descDims

def predict_preferences_given_parameters(trainingTripletArray, testingTripletArray, trainingAttrDict, testingAttrDict, lamda, dim):
    # extract categoryDescDims
    categoryDescDims = extract_desc_dims(trainingTripletArray, trainingAttrDict)
    if config.get('LogisticRegression', 'w_format') == 'concatenated':
        trainingArray = construct_cmp_results_compress_in_desc(trainingTripletArray, trainingAttrDict, categoryDescDims)
        testingArray = construct_cmp_results_in_desc(testingTripletArray, testingAttrDict, categoryDescDims)
    elif config.get('LogisticRegression', 'w_format') == 'single':
        trainingArray = construct_cmp_results_compress_in_desc_single(trainingTripletArray, trainingAttrDict, categoryDescDims)
        testingArray = construct_cmp_results_in_desc_single(testingTripletArray, testingAttrDict, categoryDescDims)
    else:
        print 'predict_preferences_given_parameters unable to recognize w_format = '+ config.get('LogisticRegression', 'w_format')
        sys.exit()
    # parameters
    normOrder = int(config.get("LogisticRegression", "norm_order"))
    maxIterations = int(config.get("LogisticRegression", "max_iterations"))
    wSize = trainingArray[0][0].size*dim
    w,convergeFlag = optimization.optimize_energy_bfgs(trainingArray, normOrder, dim, lamda*len(trainingArray)/wSize, maxIterations)
    trainingCorrect = optimization.count_hit_comparisons_compress(trainingArray, w)
    testingCorrect,resultArray = optimization.record_hit_comparisons(testingArray, w)
    return trainingCorrect,testingCorrect,convergeFlag,resultArray,w,lamda,dim

def construct_subset(tripletArray, indices):
    array = []
    for i in indices:
        array.append(tripletArray[i])
    return array

def predict_preferences_general(trainingTripletArray, testingTripletArray, trainingAttrDict, testingAttrDict):
    lamda = 0
    dim = 0
    lamda = float(config.get("LogisticRegression", "regularizer_lambda"))
    dim = int(config.get("LogisticRegression", "rank_of_w"))
    return predict_preferences_given_parameters(trainingTripletArray, testingTripletArray, trainingAttrDict, testingAttrDict,lamda, dim)


def check_overlap(array1, array2):
    for item in array1:
        if item in array2:
            print 'array1 and array2 intersects!'
            sys.exit()

def extract_one_directional_edge_training_triplets(trainingSet, testingSet):
    # extract reference and option category from testingSet
    refCateSet = {}
    optCateSet = {}
    for item in testingSet:
        refCateSet[item[4]] = 1
        optCateSet[item[5]] = 1
    print 'Ref cate set : ' + str(refCateSet)
    print 'Opt cate set : ' + str(optCateSet)
    # keep the training examples where reference in refCateSet and optition in optCateSet
    retTrainingSet = []
    for example in trainingSet:
        if (example[4] in refCateSet) and (example[5] in optCateSet):
            retTrainingSet.append(example)
    print 'extract_one_directional_edge_training_triplets get %d examples for training!'%len(retTrainingSet)
    return retTrainingSet

def extract_two_directional_edges_training_triplets(trainingSet, testingSet):
    # extract reference and option category from testingSet
    cateSet = {}
    for item in testingSet:
        cateSet[item[4]] = 1
        cateSet[item[5]] = 1
    print 'Cate set : '+str(cateSet)
    # keep the training examples where reference and option both in cateSet
    retTrainingSet = []
    for example in trainingSet:
        if (example[4] in cateSet) and (example[5] in cateSet):
            retTrainingSet.append(example)
    print 'extract_two_directional_edge_training_triplets get %d examples for training!'%len(retTrainingSet)
    return retTrainingSet

def extract_relevant_training_triplets(trainingSet, testingSet, relevantSetName):
    if relevantSetName == 'all':
        return trainingSet
    elif relevantSetName == 'one_directional_edge':
        return extract_one_directional_edge_training_triplets(trainingSet, testingSet)
    elif relevantSetName == 'two_directional_edges':
        return extract_two_directional_edges_training_triplets(trainingSet, testingSet)
    else:
        print 'extract_relevant_training_triplets unable to recognize '+relevantSetName

def extract_category_index_mapping(tripletArray):
    # [4][5] are for category names
    indexMapping = {}
    count = 0
    for triplet in tripletArray:
        if not triplet[4] in indexMapping:
            indexMapping[triplet[4]] = count
            count += 1
        if not triplet[5] in indexMapping:
            indexMapping[triplet[5]] = count
            count += 1
    return indexMapping

def replace_category_names_by_indices(tripletArray, indexMapping):
    outTripletArray = []
    for triplet in tripletArray:
        if not triplet[4] in indexMapping:
            print 'replace_category_names_by_indices uanble to find triplet[4] in indexMapping!'
            sys.exit()
        else:
            triplet[4] = indexMapping[triplet[4]]

        if not triplet[5] in indexMapping:
            print 'replace_category_names_by_indices uanble to find triplet[5] in indexMapping!'
            sys.exit()
        else:
            triplet[5] = indexMapping[triplet[5]]
        outTripletArray.append(triplet)
    return outTripletArray

# input 
#    c[0] = reference model
#    c[1] = cand1
#    c[2] = cand2
#    c[3] = relation
#    c[4]
#    c[5]
# output
#    c[0] = reference model
#    c[1] = cand1-ordered
#    c[2] = cand2-ordered
#    c[3] = {int(num of '<'), int(num of '>')}
#    c[4] (copied)
#    c[5] (copied)
def compress_training_triplet_array(tripletArray):
    dictRel = {}
    dictDim4 = {}
    dictDim5 = {}
    print 'Compress trianing triplet array in order to accelerate optimization ...'
    print '    Start with %d triplets ...'%len(tripletArray)
    for triplet in tripletArray:
        con = ''
        rel = ''
        if triplet[1] < triplet[2]:
            con = triplet[0]+'-'+triplet[1]+'-'+triplet[2]
            rel = triplet[3]
        else:
            # flip
            con = triplet[0]+'-'+triplet[2]+'-'+triplet[1]
            if triplet[3] == '<':
                rel = '>'
            elif triplet[3] == '>':
                rel = '<'
            else:
                print 'compress_training_triplet_array unable to recognize!'
                sys.exit()
        if not con in dictRel:
            dictRel[con] = []
            dictRel[con].append(0)
            dictRel[con].append(0)
            dictDim4[con] = triplet[4]
            dictDim5[con] = triplet[5]
        if rel == '<':
            dictRel[con][0] += 1
        else:
            dictRel[con][1] += 1
        if triplet[4] != dictDim4[con]:
            print 'compress_training_triplet_array dictDim4 doesnt match!'
            sys.exit()
        if triplet[5] != dictDim5[con]:
            print 'compress_training_triplet_array dictDim5 doesnt match!'
            sys.exit()
    
    tripletArrayCompress = []
    for key in dictRel:
        triplet = key.split('-')
        if len(triplet) != 3:
            print 'compress_training_triplet_array len(triplet) != 3!'
            sys.exit()
        # relation
        triplet.append(dictRel[key])
        # dim4
        if not key in dictDim4:
            print 'compress_training_triplet_array unable to find %s in dictDim4!'%key
            sys.exit()
        triplet.append(dictDim4[key])
        # dim5
        if not key in dictDim5:
            print 'compress_training_triplet_array unable to find %s in dictDim5!'%key
            sys.exit()
        triplet.append(dictDim5[key])
        tripletArrayCompress.append(triplet)
    print '    End with %d triplet groups ...'%len(tripletArrayCompress)
    return tripletArrayCompress

# prepare data
print 'Step 1.1: importing all training triplets from %s ...'%trainingTripletsDirname
# assuming trainingTripletsDirname includes a set of files, each file is a list of preferences of a single category pair
# we augment trainingCmpResults with two more dimensions, representing the reference category index and the option category index
trainingCmpResults = triplets_parser.read_multi_categories(trainingTripletsDirname)
print '    %d triplets are retrieved ...'%len(trainingCmpResults)
# import features
modelNameArray = []
for item in trainingCmpResults:
    if not item[0] in modelNameArray:
        modelNameArray.append(item[0])
    if not item[1] in modelNameArray:
        modelNameArray.append(item[1])
    if not item[2] in modelNameArray:
        modelNameArray.append(item[2])

print '    %d shapes are included ...'%(len(modelNameArray))
trainingDescriptorDict = retrieve_training_features(modelNameArray, trainingFeatureDirname)
print ''

print 'Step 1.2: importing the testing triplet ...'
testingCmpResults = triplets_parser.create_cmp_results(refModelname, opt1Modelname, opt2Modelname, refCategoryname, optCategoryname)
testingDescriptorDict = retrieve_testing_features(refSdFilename, opt1SdFilename, opt2SdFilename)
print ''

print 'Step 2: learning'
# exclude testing examples from training examples 
trainingTripletArray,testingTripletArray = pick_testing_training_examples_maximally(trainingCmpResults, testingCmpResults, 0)
# only keep training triplets according to TrainingSet::relevant_set
trainingTripletArray = extract_relevant_training_triplets(trainingTripletArray, testingTripletArray, config.get('TrainingSet', 'relevant_set'))
# to be efficient, put the different answers to the same question into the same cell
trainingTripletArrayCompress = compress_training_triplet_array(trainingTripletArray)
# the oracle accuracy on training
totalOracleCorrect = 0
totalOracleAll = 0
for item in trainingTripletArrayCompress:
    if item[3][0] > item[3][1]:
        totalOracleCorrect += item[3][0]
    else:
        totalOracleCorrect += item[3][1]
    totalOracleAll += item[3][0] + item[3][1]
# replace categoryNames by categoryIndices in trainingTripletArray
cateIndexMapping = extract_category_index_mapping(trainingTripletArray)
trainingTripletArray = replace_category_names_by_indices(trainingTripletArray, cateIndexMapping)
trainingTripletArrayCompress = replace_category_names_by_indices(trainingTripletArrayCompress, cateIndexMapping)
testingTripletArray = replace_category_names_by_indices(testingTripletArray, cateIndexMapping)


start = time.time()
trainingCorrect,testingCorrect,convergeFlag,resultArray,embeddingW,lamda,dim = predict_preferences_general(trainingTripletArrayCompress, testingTripletArray, trainingDescriptorDict, testingDescriptorDict)
end = time.time()
if convergeFlag:
    converge = 'yes'
else:
    converge = 'no'
numTestTriplets = len(testingTripletArray)
numTrainTriplets = len(trainingTripletArray)

# open log directory
os.system('mkdir -p '+workspace)
fout = open(workspace+'/log.txt', 'w')
fout.write('Test examples :')
for testingTriplet in testingTripletArray:
    fout.write('    %s,%s,%s\n'%(testingTriplet[0],testingTriplet[1],testingTriplet[2]))

fout.write('Accuracy on training: %g%%(%d/%d)\n'%(100.0*trainingCorrect/numTrainTriplets, trainingCorrect, numTrainTriplets))
#fout.write('Orcale accuracy     : %g%%(%d/%d)\n'%(100.0*totalOracleCorrect/totalOracleAll, totalOracleCorrect, totalOracleAll))

fout.write('Converge            : %s\n'%converge)

fout.write('Geom distances: ')
for idx in range(len(testingTripletArray)):
    fout.write('(%s,%s) '%(resultArray[idx][2],resultArray[idx][3]))
fout.write('\n')
fout.write('Time                : '+str(end - start)+' s\n')
fout.close()

# write embedding matrices
# index -> number of dims
descDims = extract_desc_dims(trainingTripletArrayCompress, trainingDescriptorDict)
# cateIndexMapping is name -> index 
# we get index2CateName (index -> name)
index2CateName = []
for i in descDims:
    index2CateName.append('')
for name in cateIndexMapping:
    index2CateName[cateIndexMapping[name]] = name

if config.get('LogisticRegression', 'w_format') == 'concatenated':
    startDim = 0
    endDim = 0
    embeddingW = embeddingW.T
    matrixDirname = workspace+'/matrices'
    os.system('mkdir -p '+matrixDirname)
    for index,name in enumerate(index2CateName):
        endDim = startDim + descDims[index]
        matrix = embeddingW[startDim:endDim]
        matrix = matrix.T
        # now the matrix is reducedDim by descDim
        startDim = endDim
        filename = matrixDirname+'/'+index2CateName[index]+'.txt'
        # write matrix to filename
        numpy.savetxt(filename, matrix, fmt='%g')
elif config.get('LogisticRegression', 'w_format') == 'single':
    matrixDirname = workspace+'/matrices'
    os.system('mkdir -p '+matrixDirname)
    filename = matrixDirname+'/single.txt'
    numpy.savetxt(filename, embeddingW, fmt='%g')
