# -*- coding: utf-8 -*-
"""
Created on Fri Jun  6 13:55:11 2014

@author: tialiu
"""
from numpy import *
import scipy.optimize
import utils
import scipy
import ConfigParser
#import time

config = ConfigParser.ConfigParser()
config.read('scripts/config.cfg')
# assume dist is a matrix (a column vector)
def similarity(dist, w):
    a = w*dist
    return sqrt((a.T*a).item(0))

def similarity_comparison(dist_ra, dist_rb, w):
    return similarity(dist_ra, w) - similarity(dist_rb, w)

# compute log(exp(a)+exp(b))
def log_sum_exp(a, b):
    if b > a:
        return b + log(exp(a-b)+1)
    else:
        return a + log(exp(b-a)+1)

def l_x_norm(w, norm_order):
    w_col = w.reshape((w.size, 1))
    return linalg.norm(w_col, norm_order)

# cmp_results contains a set of c, each c includes
# c[0] = distance between r and a
# c[1] = distance between r and b
# c[2] =  array of flags (1 means ra is closer than rb) [#num of ra<rb, #number of rb<ra]
# w is the dimension reduction mapping
def energy(cmp_results, lamda, norm_order, w):
    num_samples = len(cmp_results)
    # energy
    e = 0
    for i in range(num_samples):
        # comparison result
        t = cmp_results[i][2]
        # shape desc distance
        dist_ra = cmp_results[i][0]
        dist_rb = cmp_results[i][1]
        # similarity comparison
        x = similarity_comparison(dist_ra, dist_rb, w)
        # add energy
        # t[0] is the number of cases ra is closer than rb
        e = e + log_sum_exp(0, x)*t[0] + log_sum_exp(0, -x)*t[1]
    if norm_order == -1:
        # group sparsity, L1/L2 norm
        # for getting column as if getting row
        wT = w.T
        reg = 0
        for colIdx in range(w.shape[1]):
            reg += l_x_norm(wT[colIdx], 2)
        e += reg*lamda
    else:
        # L1 norm and L2 norm
        e = e + lamda*l_x_norm(w, norm_order)
    return e

def energy_likelihood(cmp_results, lamda, norm_order, w):
    num_samples = len(cmp_results)
    # energy
    e = 0
    for i in range(num_samples):
        # comparison result
        t = cmp_results[i][2]
        # shape desc distance
        dist_ra = cmp_results[i][0]
        dist_rb = cmp_results[i][1]
        # similarity comparison
        x = similarity_comparison(dist_ra, dist_rb, w)
        # add energy
        e = e + log_sum_exp(0, x)*t[0] + log_sum_exp(0, -x)*t[1]
    return e

def energy_regularization(cmp_results, lamda, norm_order, w):
    if norm_order == -1:
        # group sparsity, L1/L2 norm
        # for getting column as if getting row
        wT = w.T
        reg = 0
        for colIdx in range(w.shape[1]):
            reg += l_x_norm(wT[colIdx], 2)
        return reg*lamda
    else:
        # L1 norm and L2 norm
        return lamda*l_x_norm(w, norm_order)

def energy_finite_differential(cmp_results, lamda, norm_order, w):
    dw = matrix(zeros((w.shape[0], w.shape[1])))
    epsilon = 1e-4
    for i in range(w.shape[0]):
        for j in range(w.shape[1]):
            value = w[i,j]
            w[i,j] = value - epsilon
            energy_minus = energy(cmp_results, lamda, norm_order, w)
            w[i,j] = value + epsilon
            energy_plus = energy(cmp_results, lamda, norm_order, w)
            w[i,j] = value
            dw[i,j] = (energy_plus - energy_minus) / (2.0*epsilon)
    return dw.transpose()

# differential of energy w.r.t w (this returns dw', the transposed version)
def energy_differential(cmp_results, lamda, norm_order, w):
    num_samples = len(cmp_results)
    # matrix
    #start1 = time.time()
    major = matrix(zeros((w.shape[1], w.shape[0])))
    wT = w.T
    for i in range(num_samples):
        # shape desc distance
        dist_ra = cmp_results[i][0]
        dist_rb = cmp_results[i][1]
        # comparison result
        t = cmp_results[i][2]
        # result
        #time1 = time.time()
        emb_a = w*dist_ra
        emb_b = w*dist_rb
        sim_a = sqrt((emb_a.T*emb_a).item(0))
        sim_b = sqrt((emb_b.T*emb_b).item(0))
        #time2 = time.time()
        y = log_sum_exp(0, sim_a - sim_b)
        #time3 = time.time()
        y = exp(-y)
        #time4 = time.time()
        # (a_i*a_i^T*w^T/aa - b_i*b_i^T*w^T/bb)*(t - y)
        # aa = sqrt(a_i^T*w^T*w*a_i) = similarity(dist_ra, w)
        # bb = sqrt(b_i^T*w^T*w*b_i) = similarity(dist_rb, w)
        m1 = dist_ra*emb_a.T/sim_a
        m2 = dist_rb*emb_b.T/sim_b
        #time5 = time.time()
        major += (m1 - m2)*(t[0] - y*(t[0]+t[1]))
        #time6 = time.time()
        #print 'similarity: %g, log_sum_exp: %g, exp: %g, m-prod: %g, addition: %g'%(time2-time1, time3-time2, time4-time3, time5-time4, time6-time5)
        #raw_input('press any key to continue')
    # except for the regularization part (dw rather than dw')
    major = major.transpose()
    #end1 = time.time()
    # regularization part
    if norm_order == -1:
        wT = w.T
        # group sparsing
        # l2 for each column
        l2ForEachColInverse = empty(w.shape[1])
        for colIdx in range(w.shape[1]):
            l2ForEachColInverse[colIdx] = 1.0/l_x_norm(wT[colIdx], 2)
    else:
        l_x_norm_w_inverse = 1.0/l_x_norm(w, norm_order)
    # assume 
    for i in range(w.shape[0]):
        for j in range(w.shape[1]):
            # group sparsity: l1/l2 norm
            if norm_order == -1:
                major[i,j] += lamda*w[i,j]*l2ForEachColInverse[j]
            elif w[i,j] > 0:
                if norm_order == 1:
                    major[i,j] += lamda
                elif norm_order == 2:
                    major[i,j] += lamda*abs(w[i,j])*l_x_norm_w_inverse
                else:
                    major[i,j] += lamda*pow(abs(w[i,j])*l_x_norm_w_inverse, norm_order-1)
            elif w[i,j] < 0:
                if norm_order == 1:
                    major[i,j] -= lamda
                elif norm_order == 2:
                    major[i,j] -= lamda*abs(w[i,j])*l_x_norm_w_inverse
                else:
                    major[i,j] -= lamda*pow(abs(w[i,j])*l_x_norm_w_inverse, norm_order-1)
            # if w[i,j] = 0, don't do anything
    #print 'differential major = %g seconds'%(end1-start1)
    #print 'differential regu = %g seconds'%(end2-start2)
    #raw_input('please press any key to continue!')
    ret_major = major.transpose()
    ## debug
    #finite = energy_finite_differential(cmp_results, lamda, norm_order, w)
    #print 'dw = '+str(ret_major)
    #print 'dw(finite) = '+str(finite)
    #print 'diff = '+str(ret_major - finite)
    #print '|diff| = '+str(l_x_norm(ret_major-finite, 1))
    #raw_input('press any key to continue!')
    return ret_major

# optimize by bfgs
# wrappers
def energy_wrapper(w, cmp_results, lamda, norm_order):
    # convert w to matrix
    num_feat_dims = cmp_results[0][0].size
    reduced_num_feat_dims = len(w) / num_feat_dims
    w = matrix(w)
    w = w.reshape(reduced_num_feat_dims, num_feat_dims)
    return energy(cmp_results, lamda, norm_order, w)

def energy_differential_wrapper(w, cmp_results, lamda, norm_order):
    # matrix
    num_feat_dims = len(cmp_results[0][0])
    # convert w to matrix
    reduced_num_feat_dims = len(w) / num_feat_dims
    w = matrix(w)
    w = w.reshape(reduced_num_feat_dims, num_feat_dims)
    dw = energy_differential(cmp_results, lamda, norm_order, w).transpose()
    # conver to a 1d array
    dw = array(dw.reshape(1, dw.size))[0]
    return dw

def get_relation_from_value(distCmp):
    if distCmp < 0:
        return '<'
    elif distCmp > 0:
        return '>'
    elif random.random() < 0.5:
        return '<'
    else:
        return '>'

def is_w_consistent_to_cmp_result(w, cmp_result):
    similarity_cmp = similarity_comparison(cmp_result[0], cmp_result[1], w)
    rel = get_relation_from_value(similarity_cmp)
    if rel == '<' and cmp_result[2] > 0.5:
        return True
    elif rel == '>' and cmp_result[2] < 0.5:
        return True
    return False

def is_w_consistent_to_cmp_result_compress(w, cmp_result):
    similarity_cmp = similarity_comparison(cmp_result[0], cmp_result[1], w)
    if similarity_cmp == 0.0:
        return (cmp_result[2][0]+cmp_result[2][1])*0.5
    elif similarity_cmp < 0.0:
        return cmp_result[2][0]
    else:
        return cmp_result[2][1]

def get_flag_and_rel_and_dists(w, cmp_result):
    # repeat similarity_comparison
    dist1 = similarity(cmp_result[0], w)
    dist2 = similarity(cmp_result[1], w)
    similarity_cmp = dist1 - dist2
    rel = get_relation_from_value(similarity_cmp)
    if rel == '<' and cmp_result[2] > 0.5:
        return True,rel,str(dist1),str(dist2)
    elif rel == '>' and cmp_result[2] < 0.5:
        return True,rel,str(dist1),str(dist2)
    return False,rel,str(dist1),str(dist2)
  
def count_hit_comparisons(cmp_results, w):
    count = 0 
    for cmp_result in cmp_results:
        if is_w_consistent_to_cmp_result(w, cmp_result):
            count = count + 1
    return count

def count_hit_comparisons_compress(cmp_results, w):
    count = 0 
    for cmp_result in cmp_results:
        count += is_w_consistent_to_cmp_result_compress(w, cmp_result)
    return count


def record_hit_comparisons(cmp_results, w):
    count = 0 
    records = []
    for cmp_result in cmp_results:
        recordRow = []
        flag,rel,dist1,dist2 = get_flag_and_rel_and_dists(w, cmp_result)
        recordRow.append(rel)
        if flag:
            count = count + 1
            recordRow.append('correct')
        else:
            recordRow.append('wrong')
        recordRow.append(dist1)
        recordRow.append(dist2)
        records.append(recordRow)
    return count,records

# return a result of a converged run
def optimize_energy_bfgs(cmp_results, norm_order, reduced_num_feat_dims, lamda, maxIterations):
    # initialize a w
    num_feat_dims = cmp_results[0][0].size
    converge_flag = 0
    w_0 = random.random(reduced_num_feat_dims*num_feat_dims)
    w,fopt,gopt,Bopt,func_calls,grad_calls,warnflag,allvecs = scipy.optimize.fmin_bfgs(energy_wrapper, x0=w_0, fprime=energy_differential_wrapper, gtol=1e-6, full_output=True, maxiter=maxIterations,
                                 args=(cmp_results, lamda, norm_order, ), disp=True, retall=True)
    # convert w to matrix
    w = matrix(w)
    w = w.reshape(reduced_num_feat_dims, num_feat_dims)
    energy_final = energy(cmp_results, lamda, norm_order, w)
    if warnflag == 0:
        print 'Converge in %d iterations!'%len(allvecs)
        converge_flag = 1
    elif warnflag == 1:
        print 'Complete all %d iterations!'%len(allvecs)
    elif warnflag == 2:
        print 'Gradient/function value not changing in %d iterations!'%len(allvecs)
    else:
        print 'optimize_energy_bfgs unable to recognize warnflag = '+str(warnflag)
        sys.exit()
    energy_like = energy_likelihood(cmp_results, lamda, norm_order, w)
    energy_regu = energy_regularization(cmp_results, lamda, norm_order, w)
    return (w, converge_flag)

# cmp results contains only training examples  
# num testing runs = num of examples for cross validation
def optimize_energy_cross_validation(cmp_results, norm_order, num_testing_runs):
    maxIterations = int(config.get('LogisticRegression', 'max_iterations'))
    print 'Optimize energy cross validation ...'
    reduced_num_feat_dims_array = [1]
    lamda_array = [0.03, 0.1, 0.3, 1.0, 3.0]
    #lamda_array = [0.0]
    dim_lamda_accuracy = []
    best_correct_count = 0
    best_params = (reduced_num_feat_dims_array[0], lamda_array[0])
    best_w = matrix(random.random((reduced_num_feat_dims_array[0],cmp_results[0][0].size)))
    total_exps = len(lamda_array)*len(reduced_num_feat_dims_array)
    exp_count = 0
    # randomly select num_testing_runs indices    
    size_cmp_results = len(cmp_results)  
    index_array = []
    for i in range(size_cmp_results):
        index_array.append(i)
    random.shuffle(index_array)
    if num_testing_runs > len(index_array):
        num_testing_runs = len(index_array)
    # go through all dims
    for reduced_num_feat_dims in reduced_num_feat_dims_array:
        # go through all lambdas
        for lamda in lamda_array:
            exp_count = exp_count + 1
            print 'reduced_dim = %g, lamda = %g (%d/%d) ...'%(reduced_num_feat_dims, lamda, exp_count, total_exps)
            correct_count = 0
            # we extract one line for testing           
            utils.drawProgressBar(0)
            for i in range(num_testing_runs):
                index = index_array[i]
                # remove a sample from cmp_results
                cmp_result = cmp_results.pop(index)
                w_opt, converge_flag = optimize_energy_bfgs(cmp_results, norm_order, reduced_num_feat_dims, lamda, maxIterations)
                if is_w_consistent_to_cmp_result(w_opt, cmp_result) and converge_flag == 1:
                    correct_count = correct_count + 1
                # add the sample back to cmp_results
                cmp_results.insert(index, cmp_result)
                utils.drawProgressBar(float(i+1)/num_testing_runs)
            print " correct_count = %d/%d"%(correct_count, num_testing_runs)
            # compute w based on all input
            w, converge_flag = optimize_energy_bfgs(cmp_results, norm_order, reduced_num_feat_dims, lamda, maxIterations)
            # push to record            
            dim_lamda_accuracy.append((reduced_num_feat_dims, lamda, correct_count, converge_flag))
            # only consider if converges
            if correct_count > best_correct_count and converge_flag == 1:
                best_correct_count = correct_count
                best_params = (reduced_num_feat_dims, lamda)
                best_w = w
    print 'Print all information of cross validation (dim, lamda, correct-count)...'
    for dim_lamda_accuracy_item in dim_lamda_accuracy:
        print str(dim_lamda_accuracy_item)
    print 'Best params is '+str(best_params)
    return (best_w, best_params[0], best_params[1])
    

