import os,sys
import glob
from sklearn.decomposition import PCA
import numpy as np
# return descValueArray
# partPresenceArray
def read_all_parts_descriptor_values_for_single_shape(reponame, descTypeArray, partNameArray):
    print 'read_all_parts_descriptor_values_for_single_shape from %s ...'%reponame
    if len(descTypeArray)<1:
        print '    descTypeArray has to include at least one item!'
        sys.exit()
    # produce descValueArray and partPresenceArray
    descValueArray = []
    partPresenceArray = []
    # array< array<nparray> >
    descValueArrayPerMesh = []
    # presence
    partPresence = np.empty(0)
    for descType in descTypeArray:
        # array<nparray>
        descValueArrayPerDesc = []
        # initialize partpresence
        firstFlag = False
        if partPresence.size == 0:
            partPresence = np.zeros(len(partNameArray))
            firstFlag = True
        # check presence of the directory
        dirname = reponame+'/'+descType
        if not os.path.exists(dirname):
            print '    Unable to find %s!'%dirname
            sys.exit()
        for index,partName in enumerate(partNameArray):
            filename = dirname +'/'+partName+'.sd'
            if not os.path.exists(filename):
                descValueArrayPerDesc.append(np.empty(0))
                # set presence
                if firstFlag:
                    partPresence[index] = 0.0
                elif partPresence[index] != 0.0:
                    print '    conflicts at %s,%s'%(descType,partName)
                    sys.exit()
            else:
                fullDesc = np.loadtxt(filename)
                # one corner case, fullDesc is a number instead of a nparray
                if fullDesc.size == 1:
                    fullDesc = np.array([fullDesc])
                descValueArrayPerDesc.append(fullDesc)
                if firstFlag:
                    partPresence[index] = 1.0
                elif partPresence[index] != 1.0:
                    print '    conflicts at %s,%s'%(descType,partName)
        descValueArrayPerMesh.append(descValueArrayPerDesc)
    descValueArray.append(descValueArrayPerMesh)
    partPresenceArray.append(partPresence)
    return descValueArray,partPresenceArray

# output is array< array< array<nparray> > >
# *.sd can be found in reponame/meshname/descType
def read_all_parts_descriptor_values(reponame, descTypeArray):
    print 'read_all_parts_descriptor_values from %s ...'%reponame
    # get mesh name array and part name array
    meshNameArray = []
    partNameArray = []
    if len(descTypeArray)<1:
        print '    descTypeArray has to include at least one item!'
        sys.exit()
    for descType in descTypeArray:
        for meshDirname in glob.glob(reponame+'/*'):
            meshName = os.path.basename(meshDirname)
            if not meshName in meshNameArray:
                meshNameArray.append(meshName)
            for filename in glob.glob(meshDirname+'/'+descType+'/*.sd'):
                partName = os.path.basename(filename)
                partName = os.path.splitext(partName)[0]
                if not partName in partNameArray:
                    partNameArray.append(partName)
    print '    find %d meshes, and %d parts (%s)'%(len(meshNameArray), len(partNameArray), ','.join(partNameArray))
    # produce descValueArray and partPresenceArray
    descValueArray = []
    partPresenceArray = []
    for meshName in meshNameArray:
        # array< array<nparray> >
        descValueArrayPerMesh = []
        # presence
        partPresence = np.empty(0)
        for descType in descTypeArray:
            # array<nparray>
            descValueArrayPerDesc = []
            # initialize partpresence
            firstFlag = False
            if partPresence.size == 0:
                partPresence = np.zeros(len(partNameArray))
                firstFlag = True
            # check presence of the directory
            dirname = reponame+'/'+meshName+'/'+descType
            if not os.path.exists(dirname):
                print '    Unable to find %s!'%dirname
                sys.exit()
            for index,partName in enumerate(partNameArray):
                filename = dirname +'/'+partName+'.sd'
                if not os.path.exists(filename):
                    descValueArrayPerDesc.append(np.empty(0))
                    # set presence
                    if firstFlag:
                        partPresence[index] = 0.0
                    elif partPresence[index] != 0.0:
                        print '    conflicts at %s,%s,%s'%(meshName,descType,partName)
                        sys.exit()
                else:
                    fullDesc = np.loadtxt(filename)
                    # one corner case, fullDesc is a number instead of a nparray
                    if fullDesc.size == 1:
                        fullDesc = np.array([fullDesc])
                    descValueArrayPerDesc.append(fullDesc)
                    if firstFlag:
                        partPresence[index] = 1.0
                    elif partPresence[index] != 1.0:
                        print '    conflicts at %s,%s,%s'%(meshName,descType,partName)
            descValueArrayPerMesh.append(descValueArrayPerDesc)
        descValueArray.append(descValueArrayPerMesh)
        partPresenceArray.append(partPresence)
    return meshNameArray,partNameArray,descValueArray,partPresenceArray

def read_all_only_descriptor_values(reponame, descTypeArray):
    print 'read_all_only_descriptor_values from %s ...'%reponame
    # get mesh name array
    meshNameArray = []
    for descType in descTypeArray:
        for meshDirname in glob.glob(reponame+'/*'):
            meshName = os.path.basename(meshDirname)
            if not meshName in meshNameArray:
                meshNameArray.append(meshName)
    # get desc values
    descValueArray = []
    for meshName in meshNameArray:
        # array < array <nparray> >
        descValueArrayPerMesh = []
        for descType in descTypeArray:
            # array<nparray>
            descValueArrayPerDesc = []
            filename = reponame+'/'+meshName+'/'+descType+'/all.sd'
            if not os.path.exists(filename):
                print '    Unable to find %s!'%filename
                sys.exit()
            fullDesc = np.loadtxt(filename)
            # one corner case, fullDesc is a number instead of a nparray
            if fullDesc.size == 1:
                fullDesc = np.array([fullDesc])
            descValueArrayPerDesc.append(fullDesc)
            # descValueArrayPerDesc has only 1 element
            descValueArrayPerMesh.append(descValueArrayPerDesc)
        descValueArray.append(descValueArrayPerMesh)
    return meshNameArray,descValueArray

def write_descriptor_values_for_single_shape(newDesc, outputFilename):
    print 'Write descriptors to '+outputFilename+' ...'
    np.savetxt(outputFilename, newDesc, fmt='%g', newline=' ')

def write_descriptor_values(meshNameArray, valueArray, outputDirname):
    print 'Write descriptors to '+outputDirname+' ...'
    if len(meshNameArray) != len(valueArray):
        print '    len(meshNameArray) != len(valueArray)'
        sys.exit()
    for index,meshName in enumerate(meshNameArray):
        filename = outputDirname + '/' + meshName + '.sd'
        ## nparray seems like a column vector
        np.savetxt(filename, valueArray[index], fmt='%g', newline=' ')


# computation and filling missing
def compute_means(descValueArray, partPresenceArray):
    if len(descValueArray) < 1:
        print 'compute_means descValueArray must have more than 0 items!'
        sys.exit()
    # initialize (array< array<nparray> >)
    meanValues = []
    countValues = []
    for element in descValueArray[0]:
        # array<nparray>
        values = []
        counts = []
        for item in element:
            values.append(np.empty(0))
            counts.append(0.0)
        meanValues.append(values)
        countValues.append(counts)
    # add up
    for index,descValuesPerMesh in enumerate(descValueArray):
        partPresences = partPresenceArray[index]
        for descIdx,descValuesPerDescType in enumerate(descValuesPerMesh):
            for partIdx,descValuesPerPart in enumerate(descValuesPerDescType):
                # if this part doesn't exist
                if partPresences[partIdx] == 0.0:
                    continue
                if meanValues[descIdx][partIdx].size == 0:
                    meanValues[descIdx][partIdx] = np.array(descValuesPerPart)
                else:
                    meanValues[descIdx][partIdx] += descValuesPerPart
                countValues[descIdx][partIdx] += 1.0
    # average
    for descIdx,meanValuesPerDescType in enumerate(meanValues):
        for partIdx,meanValuesPerPart in enumerate(meanValuesPerDescType):
            if countValues[descIdx][partIdx] == 0.0:
                print 'compute_means descIdx=%d, partIdx=%d have 0 instances!'%(descIdx,partIdx)
            else:
                meanValues[descIdx][partIdx] /= countValues[descIdx][partIdx]
    return meanValues

def fill_missing_fields(descValueArray, partPresenceArray, meanValues, descTypeArray):
    for meshIdx,descValuesPerMesh in enumerate(descValueArray):
        partPresences = partPresenceArray[meshIdx]
        for descIdx,descValuesPerDescType in enumerate(descValuesPerMesh):
            for partIdx,descValuesPerPart in enumerate(descValuesPerDescType):
                # if this part doesn't exist
                if partPresences[partIdx] == 0.0:
                    # if this is area fraction, dont use mean; use 0.0 instead
                    if descTypeArray[descIdx] == 'area_fraction':
                        descValueArray[meshIdx][descIdx][partIdx] = np.array([0])
                    else:
                        descValueArray[meshIdx][descIdx][partIdx] = meanValues[descIdx][partIdx]
    return descValueArray

# concatenating, splitting and merging
# array< array< array<nparray> > > => array< array<nparray> >
def concatenate_values_for_parts(descValueArray):
    mergedValueArray = []
    for descValuesPerMesh in descValueArray:
        mergedValuesPerMesh = []
        for descValuesPerDescType in descValuesPerMesh:
            mergedValuesPerDesc = np.empty(0)
            for descValuesPerPart in descValuesPerDescType:
                #print 'mergedValuesPerDesc = '+str(mergedValuesPerDesc)
                #print 'descValuesPerPart = '+str(descValuesPerPart)
                mergedValuesPerDesc = np.concatenate((mergedValuesPerDesc, descValuesPerPart))
            mergedValuesPerMesh.append(mergedValuesPerDesc)
        mergedValueArray.append(mergedValuesPerMesh)
    return mergedValueArray

def split_based_on_pca(descValueArray, descTypeArray, descPCATypeArray):
    pcaValueArray = []
    noPcaValueArray = []
    for descValuesPerMesh in descValueArray:
        pcaValuesPerMesh = []
        noPcaValuesPerMesh = []
        if len(descTypeArray) != len(descValuesPerMesh):
            print 'split_based_on_pca len(descTypeArray) != len(descValuesPerMesh)'
            sys.exit()
        for descIdx,descType in enumerate(descTypeArray):
            if descType in descPCATypeArray:
                pcaValuesPerMesh.append(descValuesPerMesh[descIdx])
            else:
                noPcaValuesPerMesh.append(descValuesPerMesh[descIdx])
        pcaValueArray.append(pcaValuesPerMesh)
        noPcaValueArray.append(noPcaValuesPerMesh)
    return pcaValueArray,noPcaValueArray

# array< array<nparray> > + array< nparray > => array< array<nparray> >
def merge_value_mapping(noPcaValueArray, partPresenceArray):
    outputValueArray = []
    if len(noPcaValueArray) != len(partPresenceArray):
        print 'merge_value_mapping len(noPcaValueArray) != len(partPresenceArray)'
        sys.exit()
    for index, noPcaValuesPerMesh in enumerate(noPcaValueArray):
        noPcaValuesPerMesh.append(partPresenceArray[index])
        outputValueArray.append(noPcaValuesPerMesh)
    return outputValueArray

# array< array<nparray> > => array< array<nparray> > (array<nparray> only has one item)
def concatenate_values_for_desctypes(valueArray):
    outputValueArray = []
    for valuesPerMesh in valueArray:
        outputValuesPerMesh = []
        outputValues = np.empty(0)
        for valuesPerDesc in valuesPerMesh:
            outputValues = np.concatenate((outputValues,valuesPerDesc))
        outputValuesPerMesh.append(outputValues)
        outputValueArray.append(outputValuesPerMesh)
    return outputValueArray

# array< array<nparray> >+ array< array<nparray> > => array<nparray>
def merge_values(pcaValueArray, noPcaValueArray):
    if len(pcaValueArray) != len(noPcaValueArray):
        print 'merge_values len(pcaValueArray) != len(noPcaValueArray)'
        sys.exit()
    outputValueArray = []
    for index,pcaValuesPerMesh in enumerate(pcaValueArray):
        values = np.empty(0)
        noPcaValuesPerMesh = noPcaValueArray[index]
        # first no pca component
        for v in noPcaValuesPerMesh:
            values = np.concatenate((values,v))
        # second pca component
        for v in pcaValuesPerMesh:
            values = np.concatenate((values,v))
        outputValueArray.append(values)
    return outputValueArray

def merge_values_single(pcaValueArray):
    outputValueArray = []
    for index,pcaValuesPerMesh in enumerate(pcaValueArray):
        values = np.empty(0)
        for v in pcaValuesPerMesh:
            values = np.concatenate((values,v))
        outputValueArray.append(values)
    return outputValueArray

# pca
def pca_transform(testValueArray, valueArray, dims, whitenFlag):
    print 'pca transformation ...'
    if len(valueArray) == 0:
        print 'pca len(valueArray) == 0'
        sys.exit()
    nDescs = len(valueArray[0])
    if nDescs != 1:
        print 'nDescs != 1'
        sys.exit()
    stackArray = np.empty(0)
    # extract descriptors
    for valuesPerMesh in valueArray:
        desc = valuesPerMesh[0]
        if stackArray.size == 0:
            stackArray = np.empty([0, desc.size])
        stackArray = np.vstack((stackArray, desc))
    # pca transformation
    pcaTransformation = PCA(n_components=dims, whiten=whitenFlag)
    pcaTransformation.fit(stackArray)
    print '    Explained variance ratio of Descriptor '+str(pcaTransformation.explained_variance_ratio_)
    # convert testValueArray to testStackArray 
    testStackArray = np.empty(0)
    # extract descriptors
    for valuesPerMesh in testValueArray:
        desc = valuesPerMesh[0]
        if testStackArray.size == 0:
            testStackArray = np.empty([0, desc.size])
        testStackArray = np.vstack((testStackArray, desc))
    testStackArray = pcaTransformation.transform(testStackArray)
    # fill back
    for meshIdx,valuesPerMesh in enumerate(testValueArray):
        testValueArray[meshIdx][0] = testStackArray[meshIdx]
    return testValueArray

def pca(valueArray, dims, whitenFlag):
    print 'pca transformation ...'
    if len(valueArray) == 0:
        print 'pca len(valueArray) == 0'
        sys.exit()
    nDescs = len(valueArray[0])
    for descIdx in range(nDescs):
        stackArray = np.empty(0)
        # extract descriptors
        for valuesPerMesh in valueArray:
            desc = valuesPerMesh[descIdx]
            if stackArray.size == 0:
                stackArray = np.empty([0, desc.size])
            stackArray = np.vstack((stackArray, desc))
        # pca transformation
        pcaTransformation = PCA(n_components=dims, whiten=whitenFlag)
        transformedStackArray = pcaTransformation.fit_transform(stackArray)
        print '    Explained variance ratio of Descriptor %d : '%descIdx+str(pcaTransformation.explained_variance_ratio_)
        # fill back
        for meshIdx,valuesPerMesh in enumerate(valueArray):
            valueArray[meshIdx][descIdx] = transformedStackArray[meshIdx]
    return valueArray

# select the dimensions with largest variance
def select_dims_with_largest_variance(stackArray, dims):
    # subtract mean
    stackArray = stackArray - np.mean(stackArray, axis=0)
    # list by variance
    var = np.var(stackArray, axis=0)
    # extract the indices with largest variances
    indices = np.argsort(var)[::-1][:dims]
    # sub matrix
    subStackArray = stackArray[:,indices]
    # varsum
    varsum = np.sum(var)
    fraction = var[indices]/varsum
    return subStackArray,fraction

# pick_largest_variances
def remove_dims_with_small_vars(valueArray, dims):
    print 'remove_dims_with_small_vars ...'
    if len(valueArray) == 0:
        print 'remove_dims_with_small_vars len(valueArray) == 0'
        sys.exit()
    nDescs = len(valueArray[0])
    for descIdx in range(nDescs):
        stackArray = np.empty(0)
        # extract descriptors
        for valuesPerMesh in valueArray:
            desc = valuesPerMesh[descIdx]
            if stackArray.size == 0:
                stackArray = np.empty([0, desc.size])
            stackArray = np.vstack((stackArray, desc))
        # pca transformation
        transformedStackArray,fraction = select_dims_with_largest_variance(stackArray, dims)
        print '    Explained variance ratio of Descriptor %d : '%descIdx+str(fraction)
        # fill back
        for meshIdx,valuesPerMesh in enumerate(valueArray):
            valueArray[meshIdx][descIdx] = transformedStackArray[meshIdx]
    return valueArray

