# -*- coding: utf-8 -*-
"""
Created on Thu Jun  5 16:08:44 2014

@author: tialiu
"""
import os
import glob
import shape_descriptor
import csv
import optimization
import random
import numpy

# the format of triplets.csv is
# each line has name of reference shape, option1, option2, result ('>'or '<')
# and create the comparison based on a sigmoid function 1/(1+exp(dist_ra - dist_rb)),
# (this is the probability we select '<', mean dist_ra < dist_rb)
# where dist_ra = a^Tw^Twa, and a = (desc_r - desc_a)
# w is a pre-defined matrixß
def create_from_descriptors(desc_dirname, triplets_filename, w):
    comparison_array = []
    desc_filenames = glob.glob(desc_dirname+'/*.sd')
    # read descriptors from files
    shape_descriptor_dict = {}
    for desc_filename in desc_filenames:
        basename = os.path.basename(desc_filename)
        basename = os.path.splitext(basename)[0]
        shape_descriptor_dict[basename] = shape_descriptor.read(desc_filename)
    # generate all/a fraction of all triplets
    for shape_name_ref in shape_descriptor_dict:
        desc_ref = shape_descriptor_dict[shape_name_ref]
        for shape_name_opt1 in shape_descriptor_dict:
            if shape_name_opt1 == shape_name_ref:
                continue
            desc_opt1 = shape_descriptor_dict[shape_name_opt1]
            # in vector representation
            dist_ref_opt1_v = desc_ref - desc_opt1
            # in scalar representation
            dist_ref_opt1 = optimization.similarity(dist_ref_opt1_v, w)
            for shape_name_opt2 in shape_descriptor_dict:
                if shape_name_opt2 == shape_name_ref:
                    continue
                # only consider each non-identical pair
                if shape_name_opt1 >= shape_name_opt2:
                    continue
                desc_opt2 = shape_descriptor_dict[shape_name_opt2]
                # in vector representation
                dist_ref_opt2_v = desc_ref - desc_opt2
                # in scalar representation
                dist_ref_opt2 = optimization.similarity(dist_ref_opt2_v, w)
                # comparision result of a triplet
                comparison = []
                comparison.append(shape_name_ref)
                comparison.append(shape_name_opt1)
                comparison.append(shape_name_opt2)
                # a sigmoid function to decide the answer
                prob = 1.0/(1.0 + numpy.exp(dist_ref_opt1 - dist_ref_opt2))
                rand_frac = random.uniform(0, 1.0)
                if rand_frac < prob:
                    comparison.append('<')
                else:
                    comparison.append('>')
                # add to comparision_array
                comparison_array.append(comparison)
    print 'Writing to '+triplets_filename+' ...'
    fout = open(triplets_filename, 'w')
    csvwrite = csv.writer(fout)
    for triplets_comparison in comparison_array:
        csvwrite.writerow(triplets_comparison)
    fout.close()

def read(triplets_filename):
    fin = open(triplets_filename, 'r')
    csvread = csv.reader(fin)
    comparison_array = []
    for row in csvread:
        comparison_array.append(row)
    fin.close()
    return comparison_array
                    
def read_multi_categories(tripletsDirname):
    comparisonArray = []
    for tripletFilename in glob.glob(tripletsDirname+'/*.csv'):
        basename = os.path.basename(tripletFilename)
        basename = os.path.splitext(basename)[0]
        segs = basename.split('-')
        if len(segs) != 2:
            print 'trioplets_parse::read_multi_categories filename %s has the wrong format!'%tripletFilename
            sys.exit()
        refCategoryName = segs[0]
        optCategoryName = segs[1]
        fin = open(tripletFilename, 'r')
        csvread = csv.reader(fin)
        for row in csvread:
            augmentedRow = row
            augmentedRow.append(refCategoryName)
            augmentedRow.append(optCategoryName)
            comparisonArray.append(augmentedRow)
        fin.close()
    return comparisonArray

def create_cmp_results(refModelname, opt1Modelname, opt2Modelname, refCategoryname, optCategoryname):
    comparisonArray = []
    # '<' is just a place holder
    row = [refModelname, opt1Modelname, opt2Modelname, '<', refCategoryname, optCategoryname]
    comparisonArray.append(row)
    return comparisonArray

